from .buffer import _SizeConstrainedBuffer


def test_evict_oldest_items() -> None:
    """
    When eviction is necessary to make room for a new item, the oldest items
    are evicted.
    """

    buffer = _SizeConstrainedBuffer(3)

    # Fill the buffer with 3 items (each is 1 byte).
    assert buffer.add("1", b"A") is True
    assert buffer.get("1") is not None
    assert buffer.add("2", b"A") is True
    assert buffer.get("2") is not None
    assert buffer.add("3", b"A") is True
    assert buffer.get("3") is not None

    # Add a 2-byte item, which necessitates eviction.
    assert buffer.add("4", b"A" * 2) is True

    # The oldest 2 pre-existing items were evicted.
    assert buffer.get("1") is None
    assert buffer.get("2") is None

    # The most recent pre-existing item wasn't evicted.
    assert buffer.get("3") is not None

    # The new item was added.
    assert buffer.get("4") is not None


def test_item_larger_than_max_size() -> None:
    """
    If an item is larger than the max size, it is not added to the buffer.
    """

    buffer = _SizeConstrainedBuffer(1)

    assert buffer.add("1", b"A" * 2) is False
    assert buffer.get("1") is None



"""
Helper library for AI calls.

NOT STABLE! This is an experimental feature and may change in the future. If
you'd like to depend on it, we recommend copying this directory into your source
code.
"""

from . import anthropic, openai
from .base import BaseAdapter

__all__ = ["BaseAdapter", "anthropic", "openai"]



from __future__ import annotations

import inspect
import typing

from inngest._internal import (
    errors,
    execution_lib,
    function,
    net,
    server_lib,
    step_lib,
    transforms,
    types,
)

from .log import LoggerMiddleware
from .middleware import (
    Middleware,
    MiddlewareSync,
    TransformOutputResult,
    TransformOutputStepInfo,
    UninitializedMiddleware,
)

if typing.TYPE_CHECKING:
    from inngest._internal import client_lib

_mismatched_sync = errors.AsyncUnsupportedError(
    "encountered async middleware in non-async context"
)

DEFAULT_CLIENT_MIDDLEWARE: list[UninitializedMiddleware] = [LoggerMiddleware]


class MiddlewareManager:
    @property
    def middleware(self) -> list[Middleware | MiddlewareSync]:
        return [*self._middleware]

    def __init__(
        self,
        client: client_lib.Inngest,
        raw_request: object,
        timings: net.ServerTimings,
    ) -> None:
        self.client = client
        self._disabled_hooks = set[str]()
        self._middleware = list[Middleware | MiddlewareSync]()
        self._raw_request = raw_request
        self._timings = timings

    @classmethod
    def from_client(
        cls,
        client: client_lib.Inngest,
        raw_request: object,
        timings: net.ServerTimings | None,
    ) -> MiddlewareManager:
        """
        Create a new manager from an Inngest client, using the middleware on the
        client.
        """

        if timings is None:
            # A dummy timings object to simplify logic
            timings = net.ServerTimings()

        mgr = cls(client, raw_request, timings)

        for m in DEFAULT_CLIENT_MIDDLEWARE:
            mgr.add(m)

        for m in client.middleware:
            mgr.add(m)

        return mgr

    @classmethod
    def from_manager(cls, manager: MiddlewareManager) -> MiddlewareManager:
        """
        Create a new manager from another manager, using the middleware on the
        passed manager. Effectively wraps a manager.
        """
        new_mgr = cls(manager.client, manager._raw_request, manager._timings)
        for m in manager.middleware:
            new_mgr._middleware = [*new_mgr._middleware, m]
        return new_mgr

    def add(self, middleware: UninitializedMiddleware) -> None:
        self._middleware = [
            *self._middleware,
            middleware(self.client, self._raw_request),
        ]

    async def after_execution(self) -> types.MaybeError[None]:
        try:
            # Reverse order because this is an "after" hook.
            for m in reversed(self._middleware):
                await transforms.maybe_await(m.after_execution())
            return None
        except Exception as err:
            return err

    def after_execution_sync(self) -> types.MaybeError[None]:
        try:
            # Reverse order because this is an "after" hook.
            for m in reversed(self._middleware):
                if inspect.iscoroutinefunction(m.after_execution):
                    return _mismatched_sync
                m.after_execution()
            return None
        except Exception as err:
            return err

    async def after_send_events(
        self,
        result: client_lib.SendEventsResult,
    ) -> types.MaybeError[None]:
        try:
            # Reverse order because this is an "after" hook.
            for m in reversed(self._middleware):
                await transforms.maybe_await(m.after_send_events(result))
            return None
        except Exception as err:
            return err

    def after_send_events_sync(
        self,
        result: client_lib.SendEventsResult,
    ) -> types.MaybeError[None]:
        try:
            # Reverse order because this is an "after" hook.
            for m in reversed(self._middleware):
                if inspect.iscoroutinefunction(m.after_execution):
                    return _mismatched_sync
                m.after_send_events(result)
            return None
        except Exception as err:
            return err

    async def before_execution(self) -> types.MaybeError[None]:
        hook = "before_execution"
        if hook in self._disabled_hooks:
            # Only allow before_execution to be called once. This simplifies
            # code since execution can start at the function or step level.
            return None
        self._disabled_hooks.add(hook)

        try:
            for m in self._middleware:
                await transforms.maybe_await(m.before_execution())
        except Exception as err:
            return err

        return None

    def before_execution_sync(self) -> types.MaybeError[None]:
        hook = "before_execution"
        if hook in self._disabled_hooks:
            # Only allow before_execution to be called once. This simplifies
            # code since execution can start at the function or step level.
            return None
        self._disabled_hooks.add(hook)

        try:
            for m in self._middleware:
                if inspect.iscoroutinefunction(m.before_execution):
                    return _mismatched_sync
                m.before_execution()
        except Exception as err:
            return err

        return None

    async def before_response(self) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                await transforms.maybe_await(m.before_response())
            return None
        except Exception as err:
            return err

    def before_response_sync(self) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                if inspect.iscoroutinefunction(m.before_response):
                    return _mismatched_sync
                m.before_response()
            return None
        except Exception as err:
            return err

    async def before_send_events(
        self,
        events: list[server_lib.Event],
    ) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                await transforms.maybe_await(m.before_send_events(events))
            return None
        except Exception as err:
            return err

    def before_send_events_sync(
        self,
        events: list[server_lib.Event],
    ) -> types.MaybeError[None]:
        try:
            for m in self._middleware:
                if inspect.iscoroutinefunction(m.before_send_events):
                    return _mismatched_sync
                m.before_send_events(events)
            return None
        except Exception as err:
            return err

    async def transform_input(
        self,
        ctx: execution_lib.Context | execution_lib.ContextSync,
        function: function.Function[typing.Any],
        steps: step_lib.StepMemos,
    ) -> types.MaybeError[None]:
        with self._timings.mw_transform_input:
            try:
                for m in self._middleware:
                    await transforms.maybe_await(
                        m.transform_input(ctx, function, steps),
                    )
            except Exception as err:
                return err

            return None

    def transform_input_sync(
        self,
        ctx: execution_lib.Context | execution_lib.ContextSync,
        function: function.Function[typing.Any],
        steps: step_lib.StepMemos,
    ) -> types.MaybeError[None]:
        with self._timings.mw_transform_input:
            try:
                for m in self._middleware:
                    if inspect.iscoroutinefunction(m.transform_input):
                        return _mismatched_sync
                    m.transform_input(ctx, function, steps)
            except Exception as err:
                return err

            return None

    async def transform_output(
        self,
        call_res: execution_lib.CallResult,
    ) -> types.MaybeError[None]:
        with self._timings.mw_transform_output:
            # This should only happen when planning parallel steps
            if call_res.multi is not None:
                if len(call_res.multi) > 1:
                    return None
                call_res = call_res.multi[0]

            # Not sure how this can happen, but we should handle it
            if call_res.is_empty:
                return None

            # Create a new result object to pass to the middleware. We don't want to
            # pass the CallResult object because it exposes too many internal
            # implementation details
            result = TransformOutputResult(
                error=call_res.error,
                output=call_res.output,
                step=None,
            )
            if call_res.step is not None:
                result.step = TransformOutputStepInfo(
                    id=call_res.step.display_name,
                    op=call_res.step.op,
                    opts=call_res.step.opts,
                )

            try:
                # Reverse order because this is an "after" hook.
                for m in reversed(self._middleware):
                    await transforms.maybe_await(m.transform_output(result))

                # Update the original call result with the (possibly) mutated fields
                call_res.error = result.error
                call_res.output = result.output

                return None
            except Exception as err:
                return err

    def transform_output_sync(
        self,
        call_res: execution_lib.CallResult,
    ) -> types.MaybeError[None]:
        with self._timings.mw_transform_output:
            # This should only happen when planning parallel steps
            if call_res.multi is not None:
                if len(call_res.multi) > 1:
                    return None
                call_res = call_res.multi[0]

            # Not sure how this can happen, but we should handle it
            if call_res.is_empty:
                return None

            # Create a new result object to pass to the middleware. We don't want to
            # pass the CallResult object because it exposes too many internal
            # implementation details
            result = TransformOutputResult(
                error=call_res.error,
                output=call_res.output,
                step=None,
            )
            if call_res.step is not None:
                result.step = TransformOutputStepInfo(
                    id=call_res.step.display_name,
                    op=call_res.step.op,
                    opts=call_res.step.opts,
                )

            try:
                # Reverse order because this is an "after" hook.
                for m in reversed(self._middleware):
                    if isinstance(m, Middleware):
                        return _mismatched_sync
                    m.transform_output(result)

                # Update the original call result with the (possibly) mutated fields
                call_res.error = result.error
                call_res.output = result.output

                return None
            except Exception as err:
                return err



import asyncio
import inspect

import typing_extensions

from .. import execution_lib, types


def is_function_handler_async(
    value: execution_lib.FunctionHandlerAsync[types.T]
    | execution_lib.FunctionHandlerSync[types.T],
) -> typing_extensions.TypeGuard[execution_lib.FunctionHandlerAsync[types.T]]:
    return inspect.iscoroutinefunction(value)


def is_function_handler_sync(
    value: execution_lib.FunctionHandlerAsync[types.T]
    | execution_lib.FunctionHandlerSync[types.T],
) -> typing_extensions.TypeGuard[execution_lib.FunctionHandlerSync[types.T]]:
    return not inspect.iscoroutinefunction(value)


async def wait_for_next_loop() -> None:
    loop = asyncio.get_event_loop()
    fut = asyncio.Future[None]()
    loop.call_soon(lambda: fut.set_result(None))
    await fut



from inngest._internal import const, env_lib


def get_max_worker_concurrency() -> int | None:
    """
    Get the maximum number of worker concurrency from the environment.
    """
    return env_lib.get_int(const.EnvKey.CONNECT_MAX_WORKER_CONCURRENCY)



from __future__ import annotations

import dataclasses
import typing

from inngest._internal import (
    execution_lib,
    function,
    server_lib,
    step_lib,
    types,
)

if typing.TYPE_CHECKING:
    from inngest._internal import client_lib


class Middleware:
    def __init__(self, client: client_lib.Inngest, raw_request: object) -> None:
        """
        Args:
        ----
            client: Inngest client.
            raw_request: Framework/platform specific request object.
        """

        self.client = client
        self.raw_request = raw_request

    async def after_execution(self) -> None:
        """
        After executing new code. Called multiple times per run when using
        steps.
        """
        return None

    async def after_send_events(
        self,
        result: client_lib.SendEventsResult,
    ) -> None:
        """
        After sending events.
        """
        return None

    async def before_execution(self) -> None:
        """
        Before executing new code. Called multiple times per run when using
        steps.
        """
        return None

    async def before_response(self) -> None:
        """
        After the output has been set and before the response is sent
        back to Inngest. This is where you can perform any final actions before
        the response is sent back to Inngest. Called multiple times per run when
        using steps.
        """
        return None

    async def before_send_events(self, events: list[server_lib.Event]) -> None:
        """
        Before sending events.
        """
        return None

    async def transform_input(
        self,
        ctx: execution_lib.Context | execution_lib.ContextSync,
        function: function.Function[typing.Any],
        steps: step_lib.StepMemos,
    ) -> None:
        """
        Before calling a function or step. Used to replace certain arguments in
        the function. Called multiple times per run when using steps.
        """
        return None

    async def transform_output(self, result: TransformOutputResult) -> None:
        """
        After a function or step returns. Used to modify the returned data.
        Called multiple times per run when using steps. Not called when an error
        is thrown.
        """
        return None


class MiddlewareSync:
    client: client_lib.Inngest

    def __init__(self, client: client_lib.Inngest, raw_request: object) -> None:
        """
        Args:
        ----
            client: Inngest client.
            raw_request: Framework/platform specific request object.
        """

        self.client = client
        self.raw_request = raw_request

    def after_execution(self) -> None:
        """
        After executing new code. Called multiple times per run when using
        steps.
        """
        return None

    def after_send_events(
        self,
        result: client_lib.SendEventsResult,
    ) -> None:
        """
        After sending events.
        """
        return None

    def before_execution(self) -> None:
        """
        Before executing new code. Called multiple times per run when using
        steps.
        """
        return None

    def before_response(self) -> None:
        """
        After the output has been set and before the response is sent
        back to Inngest. This is where you can perform any final actions before
        the response is sent back to Inngest. Called multiple times per run when
        using steps.
        """
        return None

    def before_send_events(self, events: list[server_lib.Event]) -> None:
        """
        Before sending events.
        """
        return None

    def transform_input(
        self,
        ctx: execution_lib.Context | execution_lib.ContextSync,
        function: function.Function[typing.Any],
        steps: step_lib.StepMemos,
    ) -> None:
        """
        Before calling a function or step. Used to replace certain arguments in
        the function. Called multiple times per run when using steps.
        """
        return None

    def transform_output(self, result: TransformOutputResult) -> None:
        """
        After a function or step returns. Used to modify the returned data.
        Called multiple times per run when using steps. Not called when an error
        is thrown.
        """
        return None


UninitializedMiddleware = typing.Callable[
    # Used a "client_lib.Inngest" string to avoid a circular import
    ["client_lib.Inngest", object], Middleware | MiddlewareSync
]


@dataclasses.dataclass
class TransformOutputResult:
    # Mutations to these fields within middleware will be kept after running
    # middleware
    error: Exception | None
    output: object

    # Mutations to these fields within middleware will be discarded after
    # running middleware
    step: TransformOutputStepInfo | None

    def has_output(self) -> bool:
        return self.output is not types.empty_sentinel


@dataclasses.dataclass
class TransformOutputStepInfo:
    id: str
    op: server_lib.Opcode
    opts: dict[str, object] | None



import datetime

from . import transforms


def test_hash_signing_key() -> None:
    signing_key = "signkey-prod-568c4116828e6e8384a554153722df93022e5cd29a6c2d1b0444a19a807ff315"
    expectation = (
        "2e64ca0edc850db32ff684f967822c828f99cf57862e43205fdcf2eff8d95180"
    )
    assert transforms.hash_signing_key(signing_key) == expectation


def test_to_duration_str() -> None:
    out = transforms.to_duration_str(1000)
    assert out == "1s"

    out = transforms.to_duration_str(datetime.timedelta(minutes=2))
    assert out == "2m"



import os

from . import const, net, server_lib


def get_environment_name() -> str | None:
    """
    Get the Inngest environment name from env vars. Checks our env var and cloud
    platform (e.g. Vercel) specific env vars
    """

    return (
        os.getenv(const.EnvKey.ENV.value)
        or os.getenv(const.EnvKey.RAILWAY_GIT_BRANCH.value)
        or os.getenv(const.EnvKey.RENDER_GIT_BRANCH.value)
        or os.getenv(const.EnvKey.VERCEL_GIT_BRANCH.value)
    )


def get_url(
    env_var: const.EnvKey,
    mode: server_lib.ServerKind,
) -> str | None:
    """
    Get a URL from an env var. Returns None if the env var is not set or if its value is not a valid URL
    """

    val = os.getenv(env_var.value)
    if val is None:
        return None
    val = val.strip()

    parsed = net.parse_url(val, mode)
    if isinstance(parsed, Exception):
        return None

    return parsed


def is_false(env_var: const.EnvKey) -> bool:
    val = os.getenv(env_var.value)
    if val is None:
        return False
    val = val.strip()

    return val.lower() in ("false", "0")


def is_true(env_var: const.EnvKey) -> bool:
    val = os.getenv(env_var.value)
    if val is None:
        return False
    val = val.strip()

    return val.lower() in ("true", "1")


def is_truthy(env_var: const.EnvKey) -> bool:
    val = os.getenv(env_var.value)
    if val is None:
        return False
    val = val.strip()

    if val.lower() in ("false", "0", ""):
        return False

    return True


def get_int(env_var: const.EnvKey) -> int | None:
    """
    Get an int from an env var. Returns None if the env var is not set or if its
    value is not an int.
    """

    val = os.getenv(env_var.value)
    if not isinstance(val, str):
        return None
    try:
        return int(val)
    except ValueError:
        return None


def get_streaming(env_var: const.EnvKey) -> const.Streaming | None:
    val = os.getenv(env_var.value)
    if val is None:
        return None
    val = val.strip().lower()

    if val == "allow":
        # Accept "allow" to improve cross-language compatibility (the TS SDK has
        # "allow"). But treat it as the same as "force".
        return const.Streaming.FORCE

    try:
        return const.Streaming(val)
    except ValueError:
        return None



from __future__ import annotations

import datetime
import typing

import pydantic

from inngest._internal import const, errors, transforms, types

from .consts import DeployType, Framework
from .inspection import AuthenticatedInspection, Capabilities


class _BaseConfig(types.BaseModel):
    def convert_validation_error(
        self,
        err: pydantic.ValidationError,
    ) -> BaseException:
        return errors.FunctionConfigInvalidError.from_validation_error(err)


class Batch(_BaseConfig):
    max_size: int = pydantic.Field(..., serialization_alias="maxSize")
    timeout: int | datetime.timedelta | None = None
    key: str | None = None
    if_exp: str | None = pydantic.Field(default=None, serialization_alias="if")

    @pydantic.field_serializer("timeout")
    def serialize_timeout(
        self,
        value: int | datetime.timedelta | None,
    ) -> str | None:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class Cancel(_BaseConfig):
    event: str
    if_exp: str | None = pydantic.Field(default=None, serialization_alias="if")
    timeout: int | datetime.timedelta | None = None

    @pydantic.field_serializer("timeout")
    def serialize_timeout(
        self,
        value: int | datetime.timedelta | None,
    ) -> str | None:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class Concurrency(_BaseConfig):
    key: str | None = None
    limit: int
    scope: typing.Literal["account", "env", "fn"] | None = None


class Debounce(_BaseConfig):
    key: str | None = None
    period: int | datetime.timedelta

    @pydantic.field_serializer("period")
    def serialize_period(
        self,
        value: int | datetime.timedelta | None,
    ) -> str | None:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class FunctionConfig(_BaseConfig):
    batch_events: Batch | None = pydantic.Field(
        ..., serialization_alias="batchEvents"
    )
    cancel: list[Cancel] | None
    concurrency: list[Concurrency] | None
    debounce: Debounce | None
    id: str
    idempotency: str | None
    name: str | None
    priority: Priority | None
    rate_limit: RateLimit | None = pydantic.Field(
        ..., serialization_alias="rateLimit"
    )
    steps: dict[str, Step]
    throttle: Throttle | None
    timeouts: Timeouts | None
    singleton: Singleton | None
    triggers: list[TriggerCron | TriggerEvent]


class Priority(_BaseConfig):
    run: str


class RateLimit(_BaseConfig):
    key: str | None = None
    limit: int
    period: int | datetime.timedelta

    @pydantic.field_serializer("period")
    def serialize_period(
        self,
        value: int | datetime.timedelta | None,
    ) -> str | None:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class Retries(_BaseConfig):
    attempts: int = pydantic.Field(ge=0)


class Runtime(_BaseConfig):
    type: typing.Literal["http", "ws"]
    url: str


class Step(_BaseConfig):
    id: str
    name: str
    retries: Retries | None = None
    runtime: Runtime


class Throttle(_BaseConfig):
    key: str | None = None
    limit: int
    period: int | datetime.timedelta
    burst: int | None = 1

    @pydantic.field_serializer("period")
    def serialize_period(
        self,
        value: int | datetime.timedelta | None,
    ) -> str | None:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class Timeouts(_BaseConfig):
    start: int | datetime.timedelta | None = None
    finish: int | datetime.timedelta | None = None

    @pydantic.field_serializer("start")
    def serialize_start(
        self,
        value: int | datetime.timedelta | None,
    ) -> str | None:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out

    @pydantic.field_serializer("finish")
    def serialize_finish(
        self,
        value: int | datetime.timedelta | None,
    ) -> str | None:
        if value is None:
            return None
        out = transforms.to_duration_str(value)
        if isinstance(out, Exception):
            raise out
        return out


class Singleton(_BaseConfig):
    key: str | None = None
    mode: typing.Literal["skip", "cancel"]


class TriggerCron(_BaseConfig):
    cron: str


class TriggerEvent(_BaseConfig):
    event: str
    expression: str | None = None


class SynchronizeRequest(types.BaseModel):
    app_name: str = pydantic.Field(..., serialization_alias="appname")
    capabilities: Capabilities = Capabilities()
    deploy_type: DeployType
    framework: Framework
    functions: list[FunctionConfig] = pydantic.Field(min_length=1)
    sdk: str
    url: str
    v: str


class InBandSynchronizeRequest(types.BaseModel):
    url: str


class InBandSynchronizeResponse(types.BaseModel):
    app_id: str
    env: str | None
    framework: Framework
    functions: list[FunctionConfig]
    inspection: AuthenticatedInspection
    platform: str | None
    sdk_author: str = const.AUTHOR
    sdk_language: str = const.LANGUAGE
    sdk_version: str = const.VERSION
    url: str



"""
A helper library for realtime publishing

This is an experimental preview support for the Python SDK.
"""

from .publish import publish, publish_sync
from .subscription_tokens import (
    get_subscription_token,
    get_subscription_token_sync,
)

__all__ = [
    "publish",
    "publish_sync",
    "get_subscription_token",
    "get_subscription_token_sync",
]



import websockets

from inngest._internal import types

from . import models


async def safe_send(
    logger: types.Logger,
    state: models._State,
    ws: websockets.ClientConnection,
    message: bytes,
) -> types.MaybeError[None]:
    """
    Send a message to the WebSocket connection. If any error occurs, log and
    return the error. If the connection is closed, clear the WebSocket
    connection to trigger a reconnect.
    """

    try:
        await ws.send(message)
    except websockets.exceptions.ConnectionClosed as e:
        logger.error(f"Error sending message: {e!s}", extra={"error": str(e)})
        state.close_ws()
        return e
    except Exception as e:
        logger.error(f"Error sending message: {e!s}", extra={"error": str(e)})
        return e

    return None



from __future__ import annotations

import asyncio
import base64
import datetime
import logging
import os
import random
import secrets
import time
import typing
import urllib.parse

import httpx

from inngest._internal import (
    const,
    env_lib,
    errors,
    function,
    middleware_lib,
    net,
    serializer_lib,
    server_lib,
    types,
)

from . import models
from .utils import get_api_origin, get_event_api_origin

if typing.TYPE_CHECKING:
    from inngest._internal import execution_lib

# Dummy value
_DEV_SERVER_EVENT_KEY = "NO_EVENT_KEY_SET"

MAX_SEND_ATTEMPTS = 5
RETRY_BASE_DELAY = 0.1  # 100ms in seconds


class Inngest:
    middleware: list[middleware_lib.UninitializedMiddleware]

    @property
    def api_origin(self) -> str:
        return self._api_origin

    @property
    def env(self) -> str | None:
        return self._env

    @property
    def event_api_origin(self) -> str:
        return self._event_api_origin

    @property
    def event_key(self) -> str | None:
        return self._event_key

    @property
    def signing_key(self) -> str | None:
        return self._signing_key

    @property
    def signing_key_fallback(self) -> str | None:
        return self._signing_key_fallback

    def __init__(
        self,
        *,
        api_base_url: str | None = None,
        app_id: str,
        env: str | None = None,
        event_api_base_url: str | None = None,
        event_key: str | None = None,
        is_production: bool | None = None,
        logger: types.Logger | None = None,
        middleware: list[middleware_lib.UninitializedMiddleware] | None = None,
        request_timeout: int | datetime.timedelta | None = None,
        serializer: serializer_lib.Serializer | None = None,
        signing_key: str | None = None,
    ) -> None:
        """
        Args:
        ----
            api_base_url: Origin for the Inngest REST API.
            app_id: Unique Inngest ID. Changing this ID will make Inngest think
                it's a different app.
            env: Branch environment to use. This is only necessary for branch
                environments.
            event_api_base_url: Origin for the Inngest Event API.
            event_key: Inngest event key.
            is_production: Whether the app is in production. This affects
                request signature verification and default Inngest server URLs.
            logger: Logger to use.
            middleware: List of middleware to use.
            request_timeout: Timeout configuration for internal http client. int value is in ms. Event sending requests
                may take longer due to retries.
            serializer: Serializes/deserializes function/step output using the output_type argument.
            signing_key: Inngest signing key.
        """

        self.app_id = app_id
        self.logger = logger or logging.getLogger(__name__)
        self._mode = _get_mode(self.logger, is_production)

        # TODO: Delete this during next major version bump
        self.is_production = self._mode == server_lib.ServerKind.CLOUD

        self.middleware = middleware or []
        self._event_key = event_key or os.getenv(const.EnvKey.EVENT_KEY.value)

        self._signing_key = signing_key or os.getenv(
            const.EnvKey.SIGNING_KEY.value
        )

        self._signing_key_fallback = os.getenv(
            const.EnvKey.SIGNING_KEY_FALLBACK.value
        )

        if isinstance(request_timeout, int):
            self._httpx_timeout = request_timeout / 1000  # convert ms to s
        elif isinstance(request_timeout, datetime.timedelta):
            self._httpx_timeout = request_timeout.total_seconds()
        else:
            self._httpx_timeout = 30.0

        self._env = env or env_lib.get_environment_name()
        if (
            self._env is None
            and self._signing_key is not None
            and "branch" in self._signing_key
        ):
            self.logger.warning(
                "Signing key is for a branch environment but no branch environment is specified. This may cause unexpected behavior"
            )

        maybe_str = get_api_origin(api_base_url, self._mode)
        if isinstance(maybe_str, Exception):
            raise maybe_str
        self._api_origin = maybe_str

        maybe_str = get_event_api_origin(event_api_base_url, self._mode)
        if isinstance(maybe_str, Exception):
            raise maybe_str
        self._event_api_origin = maybe_str

        self._serializer = serializer
        self._http_client = net.AuthenticatedHTTPClient(
            env=self._env,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )

    def _build_send_request(
        self,
        events: list[server_lib.Event],
    ) -> types.MaybeError[httpx.Request]:
        event_key: str
        if self._event_key is not None:
            event_key = self._event_key
        else:
            if self._mode == server_lib.ServerKind.DEV_SERVER:
                event_key = _DEV_SERVER_EVENT_KEY
            else:
                return errors.EventKeyUnspecifiedError()

        url = urllib.parse.urljoin(self._event_api_origin, f"/e/{event_key}")

        # The client is irrespective of framework.
        framework = None

        # It'd be nice to know the expected server kind, but it's unclear how to
        # do that.
        server_kind = None

        headers = net.create_headers(
            env=self._env,
            framework=framework,
            server_kind=server_kind,
        )
        headers[server_lib.HeaderKey.EVENT_ID_SEED.value] = _seed()

        body = []
        for event in events:
            d = event.to_dict()
            if isinstance(d, Exception):
                return d

            if d.get("id") == "":
                del d["id"]
            if d.get("ts") == 0:
                d["ts"] = int(time.time() * 1000)
            body.append(d)

        return self._http_client.build_httpx_request(
            "POST", url, headers=headers, json=body, timeout=self._httpx_timeout
        )

    def add_middleware(
        self,
        middleware: middleware_lib.UninitializedMiddleware,
    ) -> None:
        self.middleware = [*self.middleware, middleware]

    def create_function(
        self,
        *,
        batch_events: server_lib.Batch | None = None,
        cancel: list[server_lib.Cancel] | None = None,
        concurrency: list[server_lib.Concurrency] | None = None,
        debounce: server_lib.Debounce | None = None,
        fn_id: str,
        idempotency: str | None = None,
        middleware: list[middleware_lib.UninitializedMiddleware] | None = None,
        name: str | None = None,
        on_failure: execution_lib.FunctionHandlerAsync[typing.Any]
        | execution_lib.FunctionHandlerSync[typing.Any]
        | None = None,
        output_type: object = types.EmptySentinel,
        priority: server_lib.Priority | None = None,
        rate_limit: server_lib.RateLimit | None = None,
        retries: int | None = None,
        throttle: server_lib.Throttle | None = None,
        timeouts: server_lib.Timeouts | None = None,
        singleton: server_lib.Singleton | None = None,
        trigger: server_lib.TriggerCron
        | server_lib.TriggerEvent
        | list[server_lib.TriggerCron | server_lib.TriggerEvent],
    ) -> typing.Callable[
        [
            execution_lib.FunctionHandlerAsync[types.T]
            | execution_lib.FunctionHandlerSync[types.T]
        ],
        function.Function[types.T],
    ]:
        """
        Create an Inngest function.

        Args:
        ----
            batch_events: Event batching config.
            cancel: Run cancellation config.
            concurrency: Concurrency config.
            debounce: Debouncing config.
            fn_id: Function ID. Changing this ID will make Inngest think this is a new function.
            idempotency: A key expression which is used to prevent duplicate events from triggering a function over 24 hours.
            middleware: Middleware to apply to this function.
            name: Human-readable function name. (Defaults to the function ID).
            on_failure: Function to call when this function fails.
            output_type: Only set if returning a non-JSON-serializable object. Related to the client's serializer argument.
            priority: Prioritize function runs.
            rate_limit: Rate limiting config.
            retries: Number of times to retry this function.
            singleton: Singleton configuration ensures that only one run per key of this function is active at any given time.
            throttle: Throttling config.
            timeouts: Timeouts config.
            trigger: What should trigger runs of this function.
        """

        fully_qualified_fn_id = f"{self.app_id}-{fn_id}"

        def decorator(
            func: execution_lib.FunctionHandlerAsync[types.T]
            | execution_lib.FunctionHandlerSync[types.T],
        ) -> function.Function[types.T]:
            triggers = trigger if isinstance(trigger, list) else [trigger]

            return function.Function(
                function.FunctionOpts(
                    batch_events=batch_events,
                    cancel=cancel,
                    concurrency=concurrency,
                    debounce=debounce,
                    fully_qualified_id=fully_qualified_fn_id,
                    idempotency=idempotency,
                    local_id=fn_id,
                    name=name or fn_id,
                    on_failure=on_failure,
                    priority=priority,
                    rate_limit=rate_limit,
                    retries=retries,
                    throttle=throttle,
                    timeouts=timeouts,
                    singleton=singleton,
                ),
                triggers,
                func,
                output_type,
                middleware,
            )

        return decorator

    async def _get_batch(
        self, run_id: str
    ) -> types.MaybeError[list[server_lib.Event]]:
        """
        Fetch a batch of events from the API
        """

        url = urllib.parse.urljoin(
            self._api_origin,
            f"/v0/runs/{run_id}/batch",
        )
        res = await self._http_client.get(url, auth=True)
        if isinstance(res, Exception):
            return res

        data = res.json()
        if not isinstance(data, list):
            raise errors.BodyInvalidError("batch data is not an array")

        events = []
        for e in data:
            events.append(server_lib.Event.model_validate(e))
        return events

    def _get_batch_sync(
        self, run_id: str
    ) -> types.MaybeError[list[server_lib.Event]]:
        """
        Fetch a batch of events from the API
        """

        url = urllib.parse.urljoin(
            self._api_origin,
            f"/v0/runs/{run_id}/batch",
        )
        res = self._http_client.get_sync(url, auth=True)
        if isinstance(res, Exception):
            return res

        data = res.json()
        if not isinstance(data, list):
            raise errors.BodyInvalidError("batch data is not an array")

        events = []
        for e in data:
            events.append(server_lib.Event.model_validate(e))
        return events

    async def _get_steps(
        self, run_id: str
    ) -> types.MaybeError[dict[str, object]]:
        """
        Fetch memoized step data from the API
        """

        url = urllib.parse.urljoin(
            self._api_origin,
            f"/v0/runs/{run_id}/actions",
        )
        res = await self._http_client.get(url, auth=True)
        if isinstance(res, Exception):
            return res

        data = res.json()
        if not isinstance(data, dict):
            raise errors.BodyInvalidError("step data is not an object")

        return data

    def _get_steps_sync(
        self, run_id: str
    ) -> types.MaybeError[dict[str, object]]:
        """
        Fetch memoized step data from the API
        """

        url = urllib.parse.urljoin(
            self._api_origin,
            f"/v0/runs/{run_id}/actions",
        )
        res = self._http_client.get_sync(url, auth=True)
        if isinstance(res, Exception):
            return res

        data = res.json()
        if not isinstance(data, dict):
            raise errors.BodyInvalidError("step data is not an object")

        return data

    async def send(
        self,
        events: server_lib.Event | list[server_lib.Event],
        *,
        skip_middleware: bool = False,
    ) -> list[str]:
        """
        Send one or more events. This method is asynchronous.

        Args:
        ----
            events: An event or list of events to send.
            skip_middleware: Whether to skip middleware.
        """

        if not isinstance(events, list):
            events = [events]

        middleware = None
        if not skip_middleware:
            middleware = middleware_lib.MiddlewareManager.from_client(
                self,
                raw_request=None,
                timings=None,
            )
            await middleware.before_send_events(events)

        req = self._build_send_request(events)
        if isinstance(req, Exception):
            raise req

        # TODO: Migrate this to HTTPClient.post
        resp = None
        for attempt in range(MAX_SEND_ATTEMPTS):
            try:
                resp = await net.fetch_with_thready_safety(
                    self._http_client._http_client,
                    self._http_client._http_client_sync,
                    req,
                )
            except httpx.RequestError:
                pass  # we will retry with delay

            # Don't retry if the request was successful or if there was a 4xx
            # status code. We don't want to retry on 4xx because the request is
            # malformed and retrying will just fail again.
            if resp is not None and resp.status_code < 500:
                break

            await asyncio.sleep(_compute_backoff_delay(attempt))

        if resp is None:
            raise errors.SendEventsError(
                "never received response while sending events", []
            )

        result = models.SendEventsResult.from_raw(resp.json())
        if isinstance(result, Exception):
            raise result

        if middleware is not None:
            err = await middleware.after_send_events(result)
            if isinstance(err, Exception):
                raise err

        if result.error is not None:
            raise errors.SendEventsError(result.error, result.ids)

        return result.ids

    def send_sync(
        self,
        events: server_lib.Event | list[server_lib.Event],
        *,
        skip_middleware: bool = False,
    ) -> list[str]:
        """
        Send one or more events. This method is synchronous.

        Args:
        ----
            events: An event or list of events to send.
            skip_middleware: Whether to skip middleware.
        """

        if not isinstance(events, list):
            events = [events]

        middleware = None
        if not skip_middleware:
            middleware = middleware_lib.MiddlewareManager.from_client(
                self,
                raw_request=None,
                timings=None,
            )
            err = middleware.before_send_events_sync(events)
            if isinstance(err, Exception):
                raise err

        req = self._build_send_request(events)
        if isinstance(req, Exception):
            raise req

        # TODO: Migrate this to HTTPClient.post_sync
        resp = None
        for attempt in range(MAX_SEND_ATTEMPTS):
            try:
                resp = self._http_client._http_client_sync.send(req)
            except httpx.RequestError:
                pass  # we will retry with delay

            # Don't retry if the request was successful or if there was a 4xx
            # status code. We don't want to retry on 4xx because the request is
            # malformed and retrying will just fail again.
            if resp is not None and resp.status_code < 500:
                break

            time.sleep(_compute_backoff_delay(attempt))

        if resp is None:
            raise errors.SendEventsError(
                "never received response while sending events", []
            )

        result = models.SendEventsResult.from_raw(resp.json())

        if isinstance(result, Exception):
            raise result

        if middleware is not None:
            err = middleware.after_send_events_sync(result)
            if isinstance(err, Exception):
                raise err

        if result.error is not None:
            raise errors.SendEventsError(result.error, result.ids)

        return result.ids

    def set_logger(self, logger: types.Logger) -> None:
        self.logger = logger

    def _serialize(self, obj: object, typ: object) -> object:
        """
        Serialize a Python object using the client's serializer.
        """

        if self._serializer is None:
            return obj

        if typ is types.EmptySentinel:
            return obj

        return self._serializer.serialize(obj, typ)

    def _deserialize(self, obj: object, typ: object) -> object:
        """
        Deserialize a Python object using the client's serializer.
        """

        if self._serializer is None:
            return obj

        if typ is types.EmptySentinel:
            return obj

        return self._serializer.deserialize(obj, typ)


def _get_mode(
    logger: types.Logger,
    is_production: bool | None,
) -> server_lib.ServerKind:
    if is_production is not None:
        if is_production:
            logger.debug("Cloud mode enabled by client argument")
            return server_lib.ServerKind.CLOUD

        logger.debug("Dev Server mode enabled by client argument")
        return server_lib.ServerKind.DEV_SERVER

    if env_lib.is_truthy(const.EnvKey.DEV):
        logger.debug(
            f"Dev Server mode enabled by {const.EnvKey.DEV.value} env var"
        )
        return server_lib.ServerKind.DEV_SERVER

    logger.debug(
        f"Cloud mode enabled. Set {const.EnvKey.DEV.value} to enable development mode"
    )
    return server_lib.ServerKind.CLOUD


def _seed() -> str:
    """
    Create the event ID seed header value. This is used to seed a
    deterministic event ID in the Inngest Server.

    Returns:
        str: A string in the format "{millis},{entropy_base64}"
    """
    current_time_millis = int(time.time() * 1000)
    entropy = secrets.token_bytes(10)
    entropy_base64 = base64.b64encode(entropy).decode("utf-8")
    return f"{current_time_millis},{entropy_base64}"


def _compute_backoff_delay(attempt: int) -> float:
    # Jitter between 0 and the base delay
    jitter = random.random() * RETRY_BASE_DELAY  # noqa:S311

    # Exponential backoff with jitter
    delay: float = RETRY_BASE_DELAY * (2**attempt) + jitter
    return delay



from __future__ import annotations

import asyncio
import datetime
import hashlib
import hmac
import http
import threading
import time
import urllib.parse

import httpx

from inngest._internal import (
    async_lib,
    config_lib,
    const,
    errors,
    server_lib,
    transforms,
    types,
)


class AuthenticatedHTTPClient:
    """
    HTTP client that:
    - Is thread-safe
    - Works in both async and sync contexts
    - Handles auth (opt in)
    - Handles branch environments
    """

    def __init__(
        self,
        *,
        env: str | None,
        request_timeout: int | datetime.timedelta | None = None,
        signing_key: str | None,
        signing_key_fallback: str | None,
    ):
        self._http_client = ThreadAwareAsyncHTTPClient().initialize()
        self._http_client_sync = httpx.Client()

        # This is probably leaking an implementation detail, and maybe we should
        # eventually remove it. In the meantime, it simplifies initial
        # HTTPClient implementation
        self.build_httpx_request = self._http_client_sync.build_request

        self._env = env
        self._signing_key = signing_key
        self._signing_key_fallback = signing_key_fallback

        if isinstance(request_timeout, int):
            self._default_timeout = request_timeout / 1000  # convert ms to s
        elif isinstance(request_timeout, datetime.timedelta):
            self._default_timeout = request_timeout.total_seconds()
        else:
            self._default_timeout = 30.0

    async def get(
        self,
        url: str,
        *,
        auth: bool = False,
        headers: dict[str, str] | None = None,
    ) -> types.MaybeError[httpx.Response]:
        """
        Perform an async HTTP GET request.

        Args:
        ----
            url: Request URL
            auth: Include the Authorization header. Never set to True if the request is not to an Inngest server
            headers: Additional request headers
        """

        req = self.build_httpx_request(
            "GET",
            url,
            headers={
                # Default headers
                **create_headers(
                    env=self._env,
                    framework=None,
                    server_kind=None,
                ),
                # Additional headers or overrides
                **(headers or {}),
            },
        )

        if auth:
            res = await fetch_with_auth_fallback(
                self._http_client,
                self._http_client_sync,
                req,
                signing_key=self._signing_key,
                signing_key_fallback=self._signing_key_fallback,
            )
        else:
            res = await fetch_with_thready_safety(
                self._http_client,
                self._http_client_sync,
                req,
            )
        if isinstance(res, Exception):
            return res

        if res.status_code >= 400:
            return Exception(f"HTTP error: {res.status_code} {res.text}")

        return res

    def get_sync(
        self,
        url: str,
        *,
        auth: bool = False,
        headers: dict[str, str] | None = None,
    ) -> types.MaybeError[httpx.Response]:
        """
        Perform a sync HTTP GET request.

        Args:
        ----
            url: Request URL
            auth: Include the Authorization header. Never set to True if the request is not to an Inngest server
            headers: Additional request headers
        """

        req = self.build_httpx_request(
            "GET",
            url,
            headers={
                # Default headers
                **create_headers(
                    env=self._env,
                    framework=None,
                    server_kind=None,
                ),
                # Additional headers or overrides
                **(headers or {}),
            },
        )

        if auth:
            res = fetch_with_auth_fallback_sync(
                self._http_client_sync,
                req,
                signing_key=self._signing_key,
                signing_key_fallback=self._signing_key_fallback,
            )
        else:
            res = self._http_client_sync.send(req)

        if isinstance(res, Exception):
            return res

        if res.status_code >= 400:
            return Exception(f"HTTP error: {res.status_code} {res.text}")

        return res

    async def post(
        self, url: str, body: object
    ) -> types.MaybeError[httpx.Response]:
        """
        Perform an asynchronous HTTP POST request. Handles authn

        Args:
        ----
            url: The pathname to the endpoint, including query string
            body: The body of the request

        Returns:
        -------
            A httpx.Response object
        """
        req = self.build_httpx_request(
            "POST",
            url,
            headers=create_headers(
                env=self._env,
                framework=None,
                server_kind=None,
            ),
            json=body,
            timeout=self._default_timeout,
        )

        res = await fetch_with_auth_fallback(
            self._http_client,
            self._http_client_sync,
            req,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if isinstance(res, Exception):
            return res

        if res.status_code >= 400:
            return Exception(f"HTTP error: {res.status_code} {res.text}")

        return res

    def post_sync(
        self, url: str, body: object
    ) -> types.MaybeError[httpx.Response]:
        """
        Perform a synchronous HTTP POST request. Handles authn

        Args:
        ----
            url: The pathname to the endpoint, including query string
            body: The body of the request

        Returns:
        -------
            A httpx.Response object
        """
        req = self.build_httpx_request(
            "POST",
            url,
            headers=create_headers(
                env=self._env,
                framework=None,
                server_kind=None,
            ),
            json=body,
            timeout=self._default_timeout,
        )

        res = fetch_with_auth_fallback_sync(
            self._http_client_sync,
            req,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if isinstance(res, Exception):
            return res

        if res.status_code >= 400:
            return Exception(f"HTTP error: {res.status_code} {res.text}")

        return res


class ThreadAwareAsyncHTTPClient(httpx.AsyncClient):
    """
    Thin wrapper around httpx.AsyncClient. It keeps track of the thread it was
    created in, which is critical since asyncio is not thread safe: calling an
    async method in a different thread will raise an exception
    """

    _creation_thread_id: int | None = None

    def is_same_thread(self) -> bool:
        if self._creation_thread_id is None:
            raise Exception("did not initialize ThreadAwareAsyncHTTPClient")

        current_thread_id = threading.get_ident()
        return self._creation_thread_id == current_thread_id

    def initialize(self) -> ThreadAwareAsyncHTTPClient:
        self._creation_thread_id = threading.get_ident()
        return self


def create_headers(
    *,
    env: str | None,
    framework: server_lib.Framework | None,
    server_kind: server_lib.ServerKind | None,
) -> dict[str, str]:
    """
    Create standard headers that should exist on every possible outgoing
    request.
    """

    headers = {
        server_lib.HeaderKey.CONTENT_TYPE.value: "application/json",
        server_lib.HeaderKey.SDK.value: f"inngest-{const.LANGUAGE}:v{const.VERSION}",
        server_lib.HeaderKey.REQUEST_VERSION.value: server_lib.PREFERRED_EXECUTION_VERSION,
        server_lib.HeaderKey.USER_AGENT.value: f"inngest-{const.LANGUAGE}:v{const.VERSION}",
    }

    if env is not None:
        headers[server_lib.HeaderKey.ENV.value] = env
    if framework is not None:
        headers[server_lib.HeaderKey.FRAMEWORK.value] = framework.value
    if server_kind is not None:
        headers[server_lib.HeaderKey.EXPECTED_SERVER_KIND.value] = (
            server_kind.value
        )

    return headers


def create_serve_url(
    *,
    public_path: str | None,
    request_url: str,
    serve_origin: str | None,
    serve_path: str | None,
) -> str:
    """
    Create the serve URL, which is the URL that the Executor will use to reach
    the SDK.

    Args:
    ----
        public_path: User-specified override for the public path.
        request_url: The URL that the Executor is using to reach the SDK.
        serve_origin: User-specified override for the serve origin.
        serve_path: User-specified override for the serve path.
    """

    # User can also specify these via env vars. The env vars take precedence.
    serve_origin = config_lib.get_serve_origin(serve_origin)
    serve_path = config_lib.get_serve_path(serve_path)

    parsed_url = urllib.parse.urlparse(request_url)
    new_scheme = parsed_url.scheme
    new_netloc = parsed_url.netloc
    new_path = parsed_url.path

    if serve_origin is not None:
        has_scheme = "://" in serve_origin
        if has_scheme:
            parsed_origin = urllib.parse.urlparse(serve_origin)
            new_scheme = parsed_origin.scheme or new_scheme
            new_netloc = parsed_origin.netloc or new_netloc
        else:
            new_scheme = "https"
            new_netloc = serve_origin

    if serve_path is not None:
        new_path = serve_path

    # public_path takes precedence over serve_path because it allows users to
    # decouple their publicly-reachable path (that the Inngest server sends
    # requests to) and the path that our SDK is hosted on. This is useful when
    # the SDK is behind a proxy that rewrites the path
    if public_path is not None:
        new_path = public_path

    return urllib.parse.urlunparse(
        (new_scheme, new_netloc, new_path, "", "", "")
    )


async def fetch_with_auth_fallback(
    client: ThreadAwareAsyncHTTPClient,
    client_sync: httpx.Client,
    request: httpx.Request,
    *,
    signing_key: str | None,
    signing_key_fallback: str | None,
) -> types.MaybeError[httpx.Response]:
    """
    Send an HTTP request with the given signing key. If the response is a 401 or
    403, then try again with the fallback signing key
    """

    if signing_key is not None:
        request.headers[server_lib.HeaderKey.AUTHORIZATION.value] = (
            f"Bearer {transforms.hash_signing_key(signing_key)}"
        )

    try:
        res = await fetch_with_thready_safety(
            client,
            client_sync,
            request,
        )
        if (
            res.status_code
            in (http.HTTPStatus.FORBIDDEN, http.HTTPStatus.UNAUTHORIZED)
            and signing_key_fallback is not None
        ):
            # Try again with the signing key fallback
            request.headers[server_lib.HeaderKey.AUTHORIZATION.value] = (
                f"Bearer {transforms.hash_signing_key(signing_key_fallback)}"
            )

            res = await fetch_with_thready_safety(
                client,
                client_sync,
                request,
            )

        return res
    except Exception as err:
        new_err = Exception(f"Failed request to {request.url}: {err}")
        new_err.__cause__ = err
        return new_err


def fetch_with_auth_fallback_sync(
    client: httpx.Client,
    request: httpx.Request,
    *,
    signing_key: str | None,
    signing_key_fallback: str | None,
) -> types.MaybeError[httpx.Response]:
    """
    Send an HTTP request with the given signing key. If the response is a 401 or
    403, then try again with the fallback signing key. Returns an error when
    receiving a non-OK response
    """

    if signing_key is not None:
        request.headers[server_lib.HeaderKey.AUTHORIZATION.value] = (
            f"Bearer {transforms.hash_signing_key(signing_key)}"
        )

    try:
        res = client.send(request)
        if (
            res.status_code
            in (http.HTTPStatus.FORBIDDEN, http.HTTPStatus.UNAUTHORIZED)
            and signing_key_fallback is not None
        ):
            # Try again with the signing key fallback
            request.headers[server_lib.HeaderKey.AUTHORIZATION.value] = (
                f"Bearer {transforms.hash_signing_key(signing_key_fallback)}"
            )
            res = client.send(request)
        return res
    except Exception as err:
        new_err = Exception(f"Failed request to {request.url}: {err}")
        new_err.__cause__ = err
        return new_err


def normalize_headers(
    headers: dict[str, str] | dict[str, list[str]],
) -> dict[str, str]:
    """
    Ensure that known headers are in the correct casing.
    """

    new_headers = {}

    for k, v in headers.items():
        k = k.lower()

        if isinstance(v, list):
            new_headers[k] = v[0]
        else:
            new_headers[k] = v

    return new_headers


def parse_url(url: str, mode: server_lib.ServerKind) -> types.MaybeError[str]:
    if "." not in url and ":" not in url.strip("http://").strip("https://"):
        return Exception("invalid URL: no domain or port")

    if url.startswith("http") is False:
        if mode is server_lib.ServerKind.CLOUD:
            url = f"https://{url}"
        else:
            url = f"http://{url}"

    try:
        parsed = urllib.parse.urlparse(url)
        if parsed.netloc == "":
            return Exception("invalid URL")

        return parsed.geturl()
    except Exception as err:
        return err


async def fetch_with_thready_safety(
    client: ThreadAwareAsyncHTTPClient,
    client_sync: httpx.Client,
    request: httpx.Request,
) -> httpx.Response:
    """
    Safely handles the situation where the async HTTP client is called in a
    different thread.
    """

    if client.is_same_thread() is True:
        # Python freaks out if you call httpx.AsyncClient's methods in multiple
        # threads. So we'll only use it if we're in the same thread as its first
        # method call
        return await client.send(request)

    loop = async_lib.get_event_loop()
    if loop is None:
        return client_sync.send(request)

    return await loop.run_in_executor(
        None,
        lambda: client_sync.send(request),
    )


def sign_request(
    body: bytes,
    signing_key: str,
    unix_ms: int | None = None,
) -> types.MaybeError[str]:
    """
    Sign an HTTP request in the same way an Inngest server would. This is only
    needed for tests that mimic Inngest server behavior.
    """

    if unix_ms is None:
        unix_ms = round(time.time())

    canonicalized = transforms.canonicalize(body)
    if isinstance(canonicalized, Exception):
        raise canonicalized

    mac = hmac.new(
        transforms.remove_signing_key_prefix(signing_key).encode("utf-8"),
        canonicalized,
        hashlib.sha256,
    )
    mac.update(str(unix_ms).encode("utf-8"))
    sig = mac.hexdigest()

    # Order matters since Inngest Cloud compares strings
    return f"t={unix_ms}&s={sig}"


def sign_response(
    body: bytes,
    signing_key: str,
    unix_ms: int | None = None,
) -> types.MaybeError[str]:
    """
    Sign an HTTP response.
    """

    if unix_ms is None:
        unix_ms = round(time.time())

    mac = hmac.new(
        transforms.remove_signing_key_prefix(signing_key).encode("utf-8"),
        body,
        hashlib.sha256,
    )
    mac.update(str(unix_ms).encode("utf-8"))
    sig = mac.hexdigest()

    # Order matters since Inngest Cloud compares strings
    return f"t={unix_ms}&s={sig}"


def _validate_sig(
    *,
    body: bytes,
    headers: dict[str, str],
    mode: server_lib.ServerKind,
    signing_key: str | None,
) -> types.MaybeError[str | None]:
    if mode == server_lib.ServerKind.DEV_SERVER:
        return None

    timestamp = None
    signature = None
    sig_header = headers.get(server_lib.HeaderKey.SIGNATURE.value)
    if sig_header is None:
        return errors.HeaderMissingError(
            f"cannot validate signature in production mode without a {server_lib.HeaderKey.SIGNATURE.value} header"
        )
    else:
        parsed = urllib.parse.parse_qs(sig_header)
        if "t" in parsed:
            timestamp = int(parsed["t"][0])
        if "s" in parsed:
            signature = parsed["s"][0]

    if signing_key is None:
        return errors.SigningKeyMissingError(
            "cannot validate signature in production mode without a signing key"
        )

    if signature is None:
        return Exception(
            f"{server_lib.HeaderKey.SIGNATURE.value} header is malformed"
        )

    mac = hmac.new(
        transforms.remove_signing_key_prefix(signing_key).encode("utf-8"),
        body,
        hashlib.sha256,
    )

    if timestamp:
        mac.update(str(timestamp).encode("utf-8"))

    if not hmac.compare_digest(signature, mac.hexdigest()):
        return errors.SigVerificationFailedError()

    return signing_key


def validate_request_sig(
    *,
    body: bytes,
    headers: dict[str, str],
    mode: server_lib.ServerKind,
    signing_key: str | None,
    signing_key_fallback: str | None,
) -> types.MaybeError[str | None]:
    """
    Validate the request signature. Falls back to the fallback signing key if
    signature validation fails with the primary signing key.

    Args:
    ----
        body: Request body.
        headers: Request headers.
        mode: Server mode.
        signing_key: Primary signing key.
        signing_key_fallback: Fallback signing key.
    """

    canonicalized = transforms.canonicalize(body)
    if isinstance(canonicalized, Exception):
        return canonicalized

    err = _validate_sig(
        body=canonicalized,
        headers=headers,
        mode=mode,
        signing_key=signing_key,
    )
    if isinstance(err, Exception) and signing_key_fallback is not None:
        # If the signature validation failed but there's a "fallback"
        # signing key, attempt to validate the signature with the fallback
        # key
        err = _validate_sig(
            body=canonicalized,
            headers=headers,
            mode=mode,
            signing_key=signing_key_fallback,
        )

    return err


def validate_response_sig(
    *,
    body: bytes,
    headers: dict[str, str],
    mode: server_lib.ServerKind,
    signing_key: str,
) -> types.MaybeError[str | None]:
    """
    Validate an HTTP response signature in the same way an Inngest server would.
    This is only needed for tests that mimic Inngest server behavior.

    Args:
    ----
        body: Request body.
        headers: Request headers.
        mode: Server mode.
        signing_key: Primary signing key.
    """

    return _validate_sig(
        body=body,
        headers=headers,
        mode=mode,
        signing_key=signing_key,
    )


class ServerTiming:
    def __init__(self, name: str) -> None:
        self._name = name
        self._start_counter: float | None = None
        self._end_counter: float | None = None

    def __enter__(self) -> ServerTiming:
        if self._start_counter is None:
            self._start_counter = time.perf_counter()
        return self

    def __exit__(self, *args: object) -> None:
        if self._end_counter is None:
            self._end_counter = time.perf_counter()

    def to_header(self) -> str:
        if self._start_counter is None or self._end_counter is None:
            return ""

        dur = int((self._end_counter - self._start_counter) * 1000)
        return f"{self._name};dur={dur}"


class _AsyncBlockServerTiming:
    """
    Special server timing that tracks how long the event loop is blocked
    """

    def __init__(self, name: str) -> None:
        self._block_dur: float = 0
        self._name = name
        self._start_counter: float | None = None
        self._tracker_task: asyncio.Task[None] | None = None

    def __enter__(self) -> _AsyncBlockServerTiming:
        if self._start_counter is None:
            self._start_counter = time.perf_counter()
        if self._tracker_task is None:
            self._tracker_task = asyncio.create_task(self._tracker())
        return self

    def __exit__(self, *args: object) -> None:
        if self._tracker_task is not None:
            self._tracker_task.cancel()

    async def _tracker(self) -> None:
        last = time.perf_counter()

        # Yield control. Without this, we'll undercount the async block duration
        # by 100 ms. It's unclear why this happens, but tests prove it
        await asyncio.sleep(0)

        interval = 0.1
        while True:
            now = time.perf_counter()
            self._block_dur += max(now - last - interval, 0)
            last = now
            await asyncio.sleep(interval)

    def to_header(self) -> str:
        if self._tracker_task is None:
            return ""

        dur = int((self._block_dur) * 1000)
        return f"{self._name};dur={dur}"


class ServerTimings:
    def __init__(self) -> None:
        # How long the event loop is blocked
        self.async_block = _AsyncBlockServerTiming("async_block")

        # CommHandler method. This should include basically everything but
        # general HTTP framework stuff (e.g. everything besides FastAPI stuff)
        self.comm_handler = ServerTiming("comm_handler")

        # Calling the Inngest function
        self.function = ServerTiming("function")

        self.mw_transform_input = ServerTiming("mw.transform_input")
        self.mw_transform_output = ServerTiming("mw.transform_output")

        # When the SDK sends an outgoing request to fetch the events and steps.
        # This happens when the incoming SDK request would be too large
        self.use_api = ServerTiming("use_api")

    def to_header(self) -> str:
        """
        Convert the server timings to the Server-Timing header value
        """

        timings: list[ServerTiming | _AsyncBlockServerTiming] = [
            self.async_block,
            self.comm_handler,
            self.function,
            self.mw_transform_input,
            self.mw_transform_output,
            self.use_api,
        ]

        # Sort by start time
        timings = sorted(
            timings,
            key=lambda x: x._start_counter or 0,
        )

        values: list[str] = [timing.to_header() for timing in timings]

        # Remove empty values
        values = [v for v in values if v != ""]

        return ", ".join(values)



import pytest

import inngest
from inngest._internal import errors


def test_sync_fn_with_async_on_failure() -> None:
    """
    Test that a sync function can not have an async on_failure handler.
    """

    client = inngest.Inngest(app_id="test", is_production=False)

    async def on_failure(ctx: inngest.Context) -> None:
        pass

    with pytest.raises(errors.Error) as e:

        @client.create_function(
            fn_id="foo",
            trigger=inngest.TriggerEvent(event="foo"),
            on_failure=on_failure,
        )
        def fn(ctx: inngest.ContextSync) -> None:
            pass

    assert (
        str(e.value)
        == "a non-async function cannot have an async on_failure handler (function foo)"
    )


def test_async_fn_with_sync_on_failure() -> None:
    """
    Test that an async function can not have a sync on_failure handler.
    """

    client = inngest.Inngest(app_id="test", is_production=False)

    def on_failure(ctx: inngest.ContextSync) -> None:
        pass

    with pytest.raises(errors.Error) as e:

        @client.create_function(
            fn_id="foo",
            trigger=inngest.TriggerEvent(event="foo"),
            on_failure=on_failure,
        )
        async def fn(ctx: inngest.Context) -> None:
            pass

    assert (
        str(e.value)
        == "an async function cannot have a non-async on_failure handler (function foo)"
    )



from __future__ import annotations

import signal
import typing

import inngest

from .connection import WorkerConnection, _WebSocketWorkerConnection


def connect(
    apps: list[tuple[inngest.Inngest, list[inngest.Function[typing.Any]]]],
    *,
    instance_id: str | None = None,
    rewrite_gateway_endpoint: typing.Callable[[str], str] | None = None,
    shutdown_signals: list[signal.Signals] | None = None,
    max_worker_concurrency: int | None = None,
) -> WorkerConnection:
    """
    Create a persistent connection to an Inngest server.

    Args:
    ----
        apps: A list of tuples, where each tuple contains an Inngest client and a list of functions.
        instance_id: A stable identifier for identifying connected apps. This can be a hostname or other identifier that remains stable across restarts. It defaults to the hostname of the machine.
        rewrite_gateway_endpoint: A function that rewrites the Inngest server gateway endpoint.
        shutdown_signals: A list of graceful shutdown signals to handle. Defaults to [SIGTERM, SIGINT].
        max_worker_concurrency: The maximum number of worker concurrency to use. Defaults to None.
    """
    return _WebSocketWorkerConnection(
        apps=apps,
        instance_id=instance_id,
        rewrite_gateway_endpoint=rewrite_gateway_endpoint,
        shutdown_signals=shutdown_signals,
        max_worker_concurrency=max_worker_concurrency,
    )



# ruff: noqa: S602, T201

import os
import pathlib
import queue
import signal
import subprocess
import threading
import time
import typing


class _CommandRunner:
    def __init__(
        self,
        command: str,
        *,
        log_path: pathlib.Path | None = None,
    ):
        self._command = command
        self._log_path = log_path
        self._process: subprocess.Popen[str] | None = None
        self._stdout_queue: queue.Queue[str] = queue.Queue()
        self._stderr_queue: queue.Queue[str] = queue.Queue()
        self._stop_printing_flag = False
        self._stdout_thread: threading.Thread | None = None
        self._stderr_thread: threading.Thread | None = None
        self._running = False

    def run(self) -> None:
        """
        Runs the command in a subprocess and starts threads for stdout and stderr (non-blocking).
        """

        if self._running:  # Prevent starting the command multiple times
            return

        self._running = True
        try:
            self._process = subprocess.Popen(
                self._command,
                preexec_fn=os.setsid,
                shell=True,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
                text=True,
            )

            self._stdout_thread = threading.Thread(
                args=(self._process.stdout, self._stdout_queue),
                target=self._enqueue_output,
            )
            self._stderr_thread = threading.Thread(
                args=(self._process.stderr, self._stderr_queue),
                target=self._enqueue_output,
            )

            self._stdout_thread.daemon = True
            self._stderr_thread.daemon = True

            self._stdout_thread.start()
            self._stderr_thread.start()

            # Start printing in a separate thread so that run is non-blocking.
            printing_thread = threading.Thread(target=self._print_output)
            printing_thread.daemon = True
            printing_thread.start()

        except FileNotFoundError:
            print(f"command not found: {self._command}")
            self._running = False
        except Exception as e:
            print(f"an error occurred: {e}")
            self._running = False

    def _enqueue_output(
        self,
        stream: typing.IO[str],
        q: queue.Queue[str],
    ) -> None:
        try:
            if self._log_path:
                with open(self._log_path, "a") as f:
                    for line in iter(stream.readline, ""):
                        q.put(line.strip())
                        f.write(line)
                        f.flush()
            else:
                for line in iter(stream.readline, ""):
                    q.put(line.strip())
        except ValueError:
            pass

    def is_running(self) -> bool:
        return self._running

    def _print_output(self) -> None:
        while self._process is not None and (
            self._process.poll() is None
            or not self._stdout_queue.empty()
            or not self._stderr_queue.empty()
        ):
            self._print_queue(self._stdout_queue)
            self._print_queue(self._stderr_queue)
            time.sleep(0.01)

        # Print any remaining output after the process finishes.
        self._print_queue(self._stdout_queue)
        self._print_queue(self._stderr_queue)
        self._running = False

    def _print_queue(self, q: queue.Queue[str]) -> None:
        while not q.empty():
            line = q.get()
            if not self._stop_printing_flag:
                print(line)

    def stop_printing(self) -> None:
        self._stop_printing_flag = True

    def wait(self) -> None:
        if self._process:
            self._process.wait()

    def kill(self) -> None:
        if self._process:
            try:
                os.killpg(os.getpgid(self._process.pid), signal.SIGTERM)
            except Exception as e:
                print(f"error killing process: {e}")
            finally:
                self._running = False



from __future__ import annotations

import logging
import typing

import pydantic
import typing_extensions

if typing.TYPE_CHECKING:
    # https://github.com/python/typeshed/issues/7855
    Logger = logging.Logger | logging.LoggerAdapter[logging.Logger]
else:
    Logger = object

T = typing.TypeVar("T")
TTuple = typing_extensions.TypeVarTuple("TTuple")


class EmptySentinel:
    pass


empty_sentinel = EmptySentinel()


# Ideally this would be recursive, but we can't do that for 2 reasons:
# 1. TypedDict is compatible with Mapping[str, object], but not anything with a
#   more restrictive value type (e.g. Mapping[str, "JSON"])
# 2. Mypy errors with "possible cyclic definition"
JSON = typing_extensions.TypeAliasType(
    "JSON",
    bool
    | int
    | float
    | str
    | typing.Mapping[str, object]
    | typing.Sequence[object]
    | None,
)


class BaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(strict=True)

    def __init__(
        __pydantic_self__,  # noqa: N805
        *args: object,
        **kwargs: object,
    ) -> None:
        try:
            super().__init__(*args, **kwargs)
        except pydantic.ValidationError as err:
            raise __pydantic_self__.convert_validation_error(err) from None

    def convert_validation_error(
        self,
        err: pydantic.ValidationError,
    ) -> BaseException:
        """
        Override this method in subclasses to convert Pydantic's
        ValidationError into a different error.
        """
        return err

    @classmethod
    def from_raw(
        cls: type[BaseModelT],
        raw: object,
    ) -> BaseModelT | Exception:
        try:
            if isinstance(raw, (str, bytes)):
                raw = cls.model_validate_json(raw)

            return cls.model_validate(raw)
        except Exception as err:
            return err

    def to_dict(self) -> MaybeError[dict[str, object]]:
        try:
            return self.model_dump(by_alias=True, mode="json")
        except Exception as err:
            return err


BaseModelT = typing.TypeVar("BaseModelT", bound=BaseModel)

MaybeError: typing_extensions.TypeAlias = T | Exception


def is_dict(v: object) -> typing.TypeGuard[dict[object, object]]:
    return isinstance(v, dict)


def is_dict_keys(
    v: object,
    key_type: type[T],
) -> typing.TypeGuard[dict[T, object]]:
    if not isinstance(v, dict):
        return False

    if key_type is not object:
        for k in typing.cast(dict[object, object], v).keys():
            if not isinstance(k, key_type):
                return False

    return True


def is_list(v: object) -> typing.TypeGuard[list[object]]:
    return isinstance(v, list)


def is_list_items(
    v: object,
    item_type: type[T],
) -> typing.TypeGuard[list[T]]:
    if not isinstance(v, list):
        return False

    if item_type is not object:
        for i in typing.cast(list[object], v):
            if not isinstance(i, item_type):
                return False

    return True



from __future__ import annotations

from inngest._internal import types

from .event import Event


class ServerRequest(types.BaseModel):
    ctx: ServerRequestCtx
    event: Event
    events: list[Event] | None = None
    steps: dict[str, object]
    use_api: bool


class ServerRequestCtx(types.BaseModel):
    attempt: int
    disable_immediate_execution: bool
    max_attempts: int | None = None
    run_id: str
    stack: ServerRequestCtxStack


class ServerRequestCtxStack(types.BaseModel):
    stack: list[str] | None = None



from __future__ import annotations

import inngest
from inngest._internal import server_lib


class Inngest(inngest.Inngest):
    """
    Mock Inngest client.
    """

    async def send(
        self,
        events: server_lib.Event | list[server_lib.Event],
        *,
        skip_middleware: bool = False,
    ) -> list[str]:
        """
        Mocked event send method.
        """

        ids = []
        for event in events:
            ids.append("00000000000000000000000000")
        return ids

    def send_sync(
        self,
        events: server_lib.Event | list[server_lib.Event],
        *,
        skip_middleware: bool = False,
    ) -> list[str]:
        """
        Mocked event send method.
        """

        ids = []
        for event in events:
            ids.append("00000000000000000000000000")
        return ids



import asyncio

from inngest._internal import types

from . import connect_pb2
from .base_handler import _BaseHandler
from .models import _State


class _DrainHandler(_BaseHandler):
    _closed_event: asyncio.Event | None = None

    def __init__(
        self,
        logger: types.Logger,
        state: _State,
    ) -> None:
        self._logger = logger
        self._state = state

    def handle_msg(
        self,
        msg: connect_pb2.ConnectMessage,
        auth_data: connect_pb2.AuthData,
        connection_id: str,
    ) -> None:
        if msg.kind != connect_pb2.GatewayMessageType.GATEWAY_CLOSING:
            return

        self._logger.debug("Draining")

        # Clear the connection to trigger reconnection logic elsewhere
        self._state.close_ws()



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GatewayMessageType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    GATEWAY_HELLO: _ClassVar[GatewayMessageType]
    WORKER_CONNECT: _ClassVar[GatewayMessageType]
    SYNC_FAILED: _ClassVar[GatewayMessageType]
    GATEWAY_CONNECTION_READY: _ClassVar[GatewayMessageType]
    GATEWAY_EXECUTOR_REQUEST: _ClassVar[GatewayMessageType]
    WORKER_READY: _ClassVar[GatewayMessageType]
    WORKER_REQUEST_ACK: _ClassVar[GatewayMessageType]
    WORKER_REQUEST_EXTEND_LEASE: _ClassVar[GatewayMessageType]
    WORKER_REQUEST_EXTEND_LEASE_ACK: _ClassVar[GatewayMessageType]
    WORKER_REPLY: _ClassVar[GatewayMessageType]
    WORKER_REPLY_ACK: _ClassVar[GatewayMessageType]
    WORKER_PAUSE: _ClassVar[GatewayMessageType]
    WORKER_HEARTBEAT: _ClassVar[GatewayMessageType]
    GATEWAY_HEARTBEAT: _ClassVar[GatewayMessageType]
    GATEWAY_CLOSING: _ClassVar[GatewayMessageType]

class SDKResponseStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    NOT_COMPLETED: _ClassVar[SDKResponseStatus]
    DONE: _ClassVar[SDKResponseStatus]
    ERROR: _ClassVar[SDKResponseStatus]

class ConnectionStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CONNECTED: _ClassVar[ConnectionStatus]
    READY: _ClassVar[ConnectionStatus]
    DRAINING: _ClassVar[ConnectionStatus]
    DISCONNECTING: _ClassVar[ConnectionStatus]
    DISCONNECTED: _ClassVar[ConnectionStatus]

class WorkerDisconnectReason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    WORKER_SHUTDOWN: _ClassVar[WorkerDisconnectReason]
    UNEXPECTED: _ClassVar[WorkerDisconnectReason]
    GATEWAY_DRAINING: _ClassVar[WorkerDisconnectReason]
    CONSECUTIVE_HEARTBEATS_MISSED: _ClassVar[WorkerDisconnectReason]
    MESSAGE_TOO_LARGE: _ClassVar[WorkerDisconnectReason]
GATEWAY_HELLO: GatewayMessageType
WORKER_CONNECT: GatewayMessageType
SYNC_FAILED: GatewayMessageType
GATEWAY_CONNECTION_READY: GatewayMessageType
GATEWAY_EXECUTOR_REQUEST: GatewayMessageType
WORKER_READY: GatewayMessageType
WORKER_REQUEST_ACK: GatewayMessageType
WORKER_REQUEST_EXTEND_LEASE: GatewayMessageType
WORKER_REQUEST_EXTEND_LEASE_ACK: GatewayMessageType
WORKER_REPLY: GatewayMessageType
WORKER_REPLY_ACK: GatewayMessageType
WORKER_PAUSE: GatewayMessageType
WORKER_HEARTBEAT: GatewayMessageType
GATEWAY_HEARTBEAT: GatewayMessageType
GATEWAY_CLOSING: GatewayMessageType
NOT_COMPLETED: SDKResponseStatus
DONE: SDKResponseStatus
ERROR: SDKResponseStatus
CONNECTED: ConnectionStatus
READY: ConnectionStatus
DRAINING: ConnectionStatus
DISCONNECTING: ConnectionStatus
DISCONNECTED: ConnectionStatus
WORKER_SHUTDOWN: WorkerDisconnectReason
UNEXPECTED: WorkerDisconnectReason
GATEWAY_DRAINING: WorkerDisconnectReason
CONSECUTIVE_HEARTBEATS_MISSED: WorkerDisconnectReason
MESSAGE_TOO_LARGE: WorkerDisconnectReason

class ConnectMessage(_message.Message):
    __slots__ = ("kind", "payload")
    KIND_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    kind: GatewayMessageType
    payload: bytes
    def __init__(self, kind: _Optional[_Union[GatewayMessageType, str]] = ..., payload: _Optional[bytes] = ...) -> None: ...

class AppConfiguration(_message.Message):
    __slots__ = ("app_name", "app_version", "functions")
    APP_NAME_FIELD_NUMBER: _ClassVar[int]
    APP_VERSION_FIELD_NUMBER: _ClassVar[int]
    FUNCTIONS_FIELD_NUMBER: _ClassVar[int]
    app_name: str
    app_version: str
    functions: bytes
    def __init__(self, app_name: _Optional[str] = ..., app_version: _Optional[str] = ..., functions: _Optional[bytes] = ...) -> None: ...

class AuthData(_message.Message):
    __slots__ = ("session_token", "sync_token")
    SESSION_TOKEN_FIELD_NUMBER: _ClassVar[int]
    SYNC_TOKEN_FIELD_NUMBER: _ClassVar[int]
    session_token: str
    sync_token: str
    def __init__(self, session_token: _Optional[str] = ..., sync_token: _Optional[str] = ...) -> None: ...

class WorkerConnectRequestData(_message.Message):
    __slots__ = ("connection_id", "instance_id", "auth_data", "capabilities", "apps", "worker_manual_readiness_ack", "system_attributes", "environment", "framework", "platform", "sdk_version", "sdk_language", "started_at", "max_worker_concurrency")
    CONNECTION_ID_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    AUTH_DATA_FIELD_NUMBER: _ClassVar[int]
    CAPABILITIES_FIELD_NUMBER: _ClassVar[int]
    APPS_FIELD_NUMBER: _ClassVar[int]
    WORKER_MANUAL_READINESS_ACK_FIELD_NUMBER: _ClassVar[int]
    SYSTEM_ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    FRAMEWORK_FIELD_NUMBER: _ClassVar[int]
    PLATFORM_FIELD_NUMBER: _ClassVar[int]
    SDK_VERSION_FIELD_NUMBER: _ClassVar[int]
    SDK_LANGUAGE_FIELD_NUMBER: _ClassVar[int]
    STARTED_AT_FIELD_NUMBER: _ClassVar[int]
    MAX_WORKER_CONCURRENCY_FIELD_NUMBER: _ClassVar[int]
    connection_id: str
    instance_id: str
    auth_data: AuthData
    capabilities: bytes
    apps: _containers.RepeatedCompositeFieldContainer[AppConfiguration]
    worker_manual_readiness_ack: bool
    system_attributes: SystemAttributes
    environment: str
    framework: str
    platform: str
    sdk_version: str
    sdk_language: str
    started_at: _timestamp_pb2.Timestamp
    max_worker_concurrency: int
    def __init__(self, connection_id: _Optional[str] = ..., instance_id: _Optional[str] = ..., auth_data: _Optional[_Union[AuthData, _Mapping]] = ..., capabilities: _Optional[bytes] = ..., apps: _Optional[_Iterable[_Union[AppConfiguration, _Mapping]]] = ..., worker_manual_readiness_ack: bool = ..., system_attributes: _Optional[_Union[SystemAttributes, _Mapping]] = ..., environment: _Optional[str] = ..., framework: _Optional[str] = ..., platform: _Optional[str] = ..., sdk_version: _Optional[str] = ..., sdk_language: _Optional[str] = ..., started_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., max_worker_concurrency: _Optional[int] = ...) -> None: ...

class GatewayConnectionReadyData(_message.Message):
    __slots__ = ("heartbeat_interval", "extend_lease_interval")
    HEARTBEAT_INTERVAL_FIELD_NUMBER: _ClassVar[int]
    EXTEND_LEASE_INTERVAL_FIELD_NUMBER: _ClassVar[int]
    heartbeat_interval: str
    extend_lease_interval: str
    def __init__(self, heartbeat_interval: _Optional[str] = ..., extend_lease_interval: _Optional[str] = ...) -> None: ...

class GatewayExecutorRequestData(_message.Message):
    __slots__ = ("request_id", "account_id", "env_id", "app_id", "app_name", "function_id", "function_slug", "step_id", "request_payload", "system_trace_ctx", "user_trace_ctx", "run_id", "lease_id")
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    ACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    ENV_ID_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    APP_NAME_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_ID_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_SLUG_FIELD_NUMBER: _ClassVar[int]
    STEP_ID_FIELD_NUMBER: _ClassVar[int]
    REQUEST_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    SYSTEM_TRACE_CTX_FIELD_NUMBER: _ClassVar[int]
    USER_TRACE_CTX_FIELD_NUMBER: _ClassVar[int]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    LEASE_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    account_id: str
    env_id: str
    app_id: str
    app_name: str
    function_id: str
    function_slug: str
    step_id: str
    request_payload: bytes
    system_trace_ctx: bytes
    user_trace_ctx: bytes
    run_id: str
    lease_id: str
    def __init__(self, request_id: _Optional[str] = ..., account_id: _Optional[str] = ..., env_id: _Optional[str] = ..., app_id: _Optional[str] = ..., app_name: _Optional[str] = ..., function_id: _Optional[str] = ..., function_slug: _Optional[str] = ..., step_id: _Optional[str] = ..., request_payload: _Optional[bytes] = ..., system_trace_ctx: _Optional[bytes] = ..., user_trace_ctx: _Optional[bytes] = ..., run_id: _Optional[str] = ..., lease_id: _Optional[str] = ...) -> None: ...

class WorkerRequestAckData(_message.Message):
    __slots__ = ("request_id", "account_id", "env_id", "app_id", "function_slug", "step_id", "system_trace_ctx", "user_trace_ctx", "run_id")
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    ACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    ENV_ID_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_SLUG_FIELD_NUMBER: _ClassVar[int]
    STEP_ID_FIELD_NUMBER: _ClassVar[int]
    SYSTEM_TRACE_CTX_FIELD_NUMBER: _ClassVar[int]
    USER_TRACE_CTX_FIELD_NUMBER: _ClassVar[int]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    account_id: str
    env_id: str
    app_id: str
    function_slug: str
    step_id: str
    system_trace_ctx: bytes
    user_trace_ctx: bytes
    run_id: str
    def __init__(self, request_id: _Optional[str] = ..., account_id: _Optional[str] = ..., env_id: _Optional[str] = ..., app_id: _Optional[str] = ..., function_slug: _Optional[str] = ..., step_id: _Optional[str] = ..., system_trace_ctx: _Optional[bytes] = ..., user_trace_ctx: _Optional[bytes] = ..., run_id: _Optional[str] = ...) -> None: ...

class WorkerRequestExtendLeaseData(_message.Message):
    __slots__ = ("request_id", "account_id", "env_id", "app_id", "function_slug", "step_id", "system_trace_ctx", "user_trace_ctx", "run_id", "lease_id")
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    ACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    ENV_ID_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_SLUG_FIELD_NUMBER: _ClassVar[int]
    STEP_ID_FIELD_NUMBER: _ClassVar[int]
    SYSTEM_TRACE_CTX_FIELD_NUMBER: _ClassVar[int]
    USER_TRACE_CTX_FIELD_NUMBER: _ClassVar[int]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    LEASE_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    account_id: str
    env_id: str
    app_id: str
    function_slug: str
    step_id: str
    system_trace_ctx: bytes
    user_trace_ctx: bytes
    run_id: str
    lease_id: str
    def __init__(self, request_id: _Optional[str] = ..., account_id: _Optional[str] = ..., env_id: _Optional[str] = ..., app_id: _Optional[str] = ..., function_slug: _Optional[str] = ..., step_id: _Optional[str] = ..., system_trace_ctx: _Optional[bytes] = ..., user_trace_ctx: _Optional[bytes] = ..., run_id: _Optional[str] = ..., lease_id: _Optional[str] = ...) -> None: ...

class WorkerRequestExtendLeaseAckData(_message.Message):
    __slots__ = ("request_id", "account_id", "env_id", "app_id", "function_slug", "new_lease_id")
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    ACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    ENV_ID_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    FUNCTION_SLUG_FIELD_NUMBER: _ClassVar[int]
    NEW_LEASE_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    account_id: str
    env_id: str
    app_id: str
    function_slug: str
    new_lease_id: str
    def __init__(self, request_id: _Optional[str] = ..., account_id: _Optional[str] = ..., env_id: _Optional[str] = ..., app_id: _Optional[str] = ..., function_slug: _Optional[str] = ..., new_lease_id: _Optional[str] = ...) -> None: ...

class SDKResponse(_message.Message):
    __slots__ = ("request_id", "account_id", "env_id", "app_id", "status", "body", "no_retry", "retry_after", "sdk_version", "request_version", "system_trace_ctx", "user_trace_ctx", "run_id")
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    ACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    ENV_ID_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    BODY_FIELD_NUMBER: _ClassVar[int]
    NO_RETRY_FIELD_NUMBER: _ClassVar[int]
    RETRY_AFTER_FIELD_NUMBER: _ClassVar[int]
    SDK_VERSION_FIELD_NUMBER: _ClassVar[int]
    REQUEST_VERSION_FIELD_NUMBER: _ClassVar[int]
    SYSTEM_TRACE_CTX_FIELD_NUMBER: _ClassVar[int]
    USER_TRACE_CTX_FIELD_NUMBER: _ClassVar[int]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    account_id: str
    env_id: str
    app_id: str
    status: SDKResponseStatus
    body: bytes
    no_retry: bool
    retry_after: str
    sdk_version: str
    request_version: int
    system_trace_ctx: bytes
    user_trace_ctx: bytes
    run_id: str
    def __init__(self, request_id: _Optional[str] = ..., account_id: _Optional[str] = ..., env_id: _Optional[str] = ..., app_id: _Optional[str] = ..., status: _Optional[_Union[SDKResponseStatus, str]] = ..., body: _Optional[bytes] = ..., no_retry: bool = ..., retry_after: _Optional[str] = ..., sdk_version: _Optional[str] = ..., request_version: _Optional[int] = ..., system_trace_ctx: _Optional[bytes] = ..., user_trace_ctx: _Optional[bytes] = ..., run_id: _Optional[str] = ...) -> None: ...

class WorkerReplyAckData(_message.Message):
    __slots__ = ("request_id",)
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    def __init__(self, request_id: _Optional[str] = ...) -> None: ...

class ConnMetadata(_message.Message):
    __slots__ = ("id", "gateway_id", "instance_id", "all_worker_groups", "synced_worker_groups", "status", "last_heartbeat_at", "sdk_language", "sdk_version", "attributes")
    class AllWorkerGroupsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class SyncedWorkerGroupsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    ID_FIELD_NUMBER: _ClassVar[int]
    GATEWAY_ID_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_ID_FIELD_NUMBER: _ClassVar[int]
    ALL_WORKER_GROUPS_FIELD_NUMBER: _ClassVar[int]
    SYNCED_WORKER_GROUPS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    LAST_HEARTBEAT_AT_FIELD_NUMBER: _ClassVar[int]
    SDK_LANGUAGE_FIELD_NUMBER: _ClassVar[int]
    SDK_VERSION_FIELD_NUMBER: _ClassVar[int]
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    id: str
    gateway_id: str
    instance_id: str
    all_worker_groups: _containers.ScalarMap[str, str]
    synced_worker_groups: _containers.ScalarMap[str, str]
    status: ConnectionStatus
    last_heartbeat_at: _timestamp_pb2.Timestamp
    sdk_language: str
    sdk_version: str
    attributes: SystemAttributes
    def __init__(self, id: _Optional[str] = ..., gateway_id: _Optional[str] = ..., instance_id: _Optional[str] = ..., all_worker_groups: _Optional[_Mapping[str, str]] = ..., synced_worker_groups: _Optional[_Mapping[str, str]] = ..., status: _Optional[_Union[ConnectionStatus, str]] = ..., last_heartbeat_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., sdk_language: _Optional[str] = ..., sdk_version: _Optional[str] = ..., attributes: _Optional[_Union[SystemAttributes, _Mapping]] = ...) -> None: ...

class SystemAttributes(_message.Message):
    __slots__ = ("cpu_cores", "mem_bytes", "os")
    CPU_CORES_FIELD_NUMBER: _ClassVar[int]
    MEM_BYTES_FIELD_NUMBER: _ClassVar[int]
    OS_FIELD_NUMBER: _ClassVar[int]
    cpu_cores: int
    mem_bytes: int
    os: str
    def __init__(self, cpu_cores: _Optional[int] = ..., mem_bytes: _Optional[int] = ..., os: _Optional[str] = ...) -> None: ...

class ConnGroup(_message.Message):
    __slots__ = ("env_id", "app_id", "app_name", "hash", "conns", "sync_id", "app_version")
    ENV_ID_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    APP_NAME_FIELD_NUMBER: _ClassVar[int]
    HASH_FIELD_NUMBER: _ClassVar[int]
    CONNS_FIELD_NUMBER: _ClassVar[int]
    SYNC_ID_FIELD_NUMBER: _ClassVar[int]
    APP_VERSION_FIELD_NUMBER: _ClassVar[int]
    env_id: str
    app_id: str
    app_name: str
    hash: str
    conns: _containers.RepeatedCompositeFieldContainer[ConnMetadata]
    sync_id: str
    app_version: str
    def __init__(self, env_id: _Optional[str] = ..., app_id: _Optional[str] = ..., app_name: _Optional[str] = ..., hash: _Optional[str] = ..., conns: _Optional[_Iterable[_Union[ConnMetadata, _Mapping]]] = ..., sync_id: _Optional[str] = ..., app_version: _Optional[str] = ...) -> None: ...

class StartResponse(_message.Message):
    __slots__ = ("connection_id", "gateway_endpoint", "gateway_group", "session_token", "sync_token")
    CONNECTION_ID_FIELD_NUMBER: _ClassVar[int]
    GATEWAY_ENDPOINT_FIELD_NUMBER: _ClassVar[int]
    GATEWAY_GROUP_FIELD_NUMBER: _ClassVar[int]
    SESSION_TOKEN_FIELD_NUMBER: _ClassVar[int]
    SYNC_TOKEN_FIELD_NUMBER: _ClassVar[int]
    connection_id: str
    gateway_endpoint: str
    gateway_group: str
    session_token: str
    sync_token: str
    def __init__(self, connection_id: _Optional[str] = ..., gateway_endpoint: _Optional[str] = ..., gateway_group: _Optional[str] = ..., session_token: _Optional[str] = ..., sync_token: _Optional[str] = ...) -> None: ...

class StartRequest(_message.Message):
    __slots__ = ("exclude_gateways",)
    EXCLUDE_GATEWAYS_FIELD_NUMBER: _ClassVar[int]
    exclude_gateways: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, exclude_gateways: _Optional[_Iterable[str]] = ...) -> None: ...

class FlushResponse(_message.Message):
    __slots__ = ("request_id",)
    REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    request_id: str
    def __init__(self, request_id: _Optional[str] = ...) -> None: ...

class PubSubAckMessage(_message.Message):
    __slots__ = ("ts", "nack", "nack_reason")
    TS_FIELD_NUMBER: _ClassVar[int]
    NACK_FIELD_NUMBER: _ClassVar[int]
    NACK_REASON_FIELD_NUMBER: _ClassVar[int]
    ts: _timestamp_pb2.Timestamp
    nack: bool
    nack_reason: SystemError
    def __init__(self, ts: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., nack: bool = ..., nack_reason: _Optional[_Union[SystemError, _Mapping]] = ...) -> None: ...

class SystemError(_message.Message):
    __slots__ = ("code", "data", "message")
    CODE_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    code: str
    data: bytes
    message: str
    def __init__(self, code: _Optional[str] = ..., data: _Optional[bytes] = ..., message: _Optional[str] = ...) -> None: ...



import typing
from urllib.parse import urlencode, urljoin

from inngest._internal import client_lib


# TODO - Support streams
async def publish(
    client: client_lib.Inngest,
    channel: str,
    topic: str,
    data: typing.Mapping[str, object],
) -> None:
    """
    Publish a message to a realtime channel.

    Args:
    ----
        client: The Inngest client
        channel: The realtime channel name
        topic: The realtime topic name
        data: JSON-serializable data to publish to subscribers

    Raises:
    ------
        errors.Error: If data is not JSON serializable or if publishing fails
    """

    params = {
        "channel": channel,
        "topic": topic,
    }

    res = await client._http_client.post(
        url=urljoin(
            client._api_origin, f"/v1/realtime/publish?{urlencode(params)}"
        ),
        body=data,
    )
    if isinstance(res, Exception):
        raise res
    return None


def publish_sync(
    client: client_lib.Inngest,
    channel: str,
    topic: str,
    data: typing.Mapping[str, object],
) -> None:
    """
    Publish a message to a realtime channel synchronously.

    Args:
    ----
        client: The Inngest client
        channel: The realtime channel name
        topic: The realtime topic name
        data: JSON-serializable data to publish to subscribers

    Raises:
    ------
        errors.Error: If data is not JSON serializable or if publishing fails
    """

    params = {
        "channel": channel,
        "topic": topic,
    }

    res = client._http_client.post_sync(
        url=urljoin(
            client._api_origin, f"/v1/realtime/publish?{urlencode(params)}"
        ),
        body=data,
    )
    if isinstance(res, Exception):
        raise res
    return None



from .base import BaseExecution, BaseExecutionSync
from .models import (
    CallResult,
    Context,
    ContextSync,
    FunctionHandlerAsync,
    FunctionHandlerSync,
    ReportedStep,
    UserError,
)
from .utils import is_function_handler_async, is_function_handler_sync
from .v0 import ExecutionV0, ExecutionV0Sync

__all__ = [
    "BaseExecution",
    "BaseExecutionSync",
    "CallResult",
    "Context",
    "ContextSync",
    "FunctionHandlerAsync",
    "FunctionHandlerSync",
    "ExecutionV0",
    "ExecutionV0Sync",
    "ReportedStep",
    "UserError",
    "is_function_handler_async",
    "is_function_handler_sync",
]



import asyncio
import typing
import unittest

import pytest

from .value_watcher import _ValueWatcher


class TestValueWatcher(unittest.IsolatedAsyncioTestCase):
    async def _start_tasks(
        self, *coros: typing.Coroutine[typing.Any, typing.Any, typing.Any]
    ) -> typing.Callable[[], typing.Awaitable[None]]:
        tasks = [asyncio.create_task(coro) for coro in coros]
        for task in tasks:
            self.addCleanup(task.cancel)

        # Allow an event loop iteration to pass, ensuring that the tasks start.
        await asyncio.sleep(0)

        async def wait() -> None:
            await asyncio.gather(*tasks)

        return wait

    @pytest.mark.timeout(2, method="thread")
    async def test_wait_for(self) -> None:
        watcher = _ValueWatcher(0)

        # Immediately returns because the value is already 0.
        await watcher.wait_for(0)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for(1),
        )

        watcher.value = 1

        # Wait for all watchers to complete.
        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_wait_for_not(self) -> None:
        watcher = _ValueWatcher(0)

        # Immediately returns because the value is already 0.
        await watcher.wait_for_not(1)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for_not(0),
        )

        watcher.value = 1

        # Wait for all watchers to complete.
        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_wait_for_change(self) -> None:
        watcher = _ValueWatcher(0)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for_change(),
        )

        watcher.value = 1

        # Wait for all watchers to complete.
        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_watch_wait_for_and_not(self) -> None:
        """
        Match parallel watchers:
        - Wait for the value to not be a certain value.
        - Wait for the value to be a certain value.
        """

        watcher = _ValueWatcher(0)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for_not(0),
            watcher.wait_for(1),
        )

        # Set the value, which should propagate to all watchers.
        watcher.value = 1

        # Wait for all watchers to complete.
        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_quickly_set_twice(self) -> None:
        """
        Setting twice really quickly should not cause a missed value.
        """

        watcher = _ValueWatcher(0)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for(1),
            watcher.wait_for(2),
        )

        # Set the value, which should propagate to all watchers.
        watcher.value = 1
        watcher.value = 2

        # Wait for all watchers to complete.
        await wait_for_tasks()



from __future__ import annotations

from .base import BaseAdapter


class Adapter(BaseAdapter):
    """
    Anthropic adapter
    """

    def __init__(
        self,
        *,
        auth_key: str,
        base_url: str | None = None,
        headers: dict[str, str] | None = None,
        model: str,
    ) -> None:
        """
        Args:
        ----
            auth_key: Anthropic API key.
            base_url: Anthropic API URL.
            headers: Additional headers to send with the request.
            model: Anthropic model to use.
        """

        self._auth_key = auth_key
        self._headers = headers or {"anthropic-version": "2023-06-01"}
        self._model = model
        self._url = base_url or "https://api.anthropic.com/v1"

    def auth_key(self) -> str:
        """
        Return the authentication key for the adapter.
        """

        return self._auth_key

    def format(self) -> str:
        """
        Return the format for the adapter.
        """

        return "anthropic"

    def headers(self) -> dict[str, str]:
        """
        Return the headers for the adapter.
        """

        return self._headers

    def on_call(self, body: dict[str, object]) -> None:
        """
        Modify the request body.
        """

        if not body.get("model"):
            body["model"] = self._model

    def url_infer(self) -> str:
        """
        Return the URL for generating text.
        """

        return self._url.rstrip("/") + "/messages"



import datetime
import typing
import unittest

import pytest

import inngest
from inngest.experimental import mocked

from .errors import UnstubbedStepError

client = inngest.Inngest(app_id="my-app")
client_mock = mocked.Inngest(app_id="test")


class TestTriggerAsync(unittest.TestCase):
    def test_parallel(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        async def fn(ctx: inngest.Context) -> tuple[str, ...]:
            async def foo(value: str) -> str:
                return value

            return await ctx.group.parallel(
                (
                    lambda: ctx.step.run("a", foo, "a"),
                    lambda: ctx.step.run("b", foo, "b"),
                )
            )

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.COMPLETED
        assert res.output == ("a", "b")


class TestTriggerSync(unittest.TestCase):
    def test_no_steps(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> str:
            return "hi"

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.COMPLETED
        assert res.output == "hi"

    def test_two_steps(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> str:
            ctx.step.run("a", lambda: None)
            ctx.step.run("b", lambda: None)
            return "hi"

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.COMPLETED
        assert res.output == "hi"

    def test_client_send(self) -> None:
        """
        TODO: Figure out how to support this use case. Since the client in the
        Inngest function is real, it's trying to send the event to a real
        Inngest server.
        """

        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> list[str]:
            return client.send_sync(
                [
                    inngest.Event(name="other-event"),
                    inngest.Event(name="other-event"),
                ]
            )

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.FAILED

    def test_send_event(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> list[str]:
            return ctx.step.send_event(
                "a",
                [
                    inngest.Event(name="event-1"),
                    inngest.Event(name="event-2"),
                ],
            )

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.COMPLETED
        assert res.output == [
            "00000000000000000000000000",
            "00000000000000000000000000",
        ]

    def test_invoke(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> typing.Any:
            return ctx.step.invoke_by_id(
                "a",
                app_id="foo",
                function_id="bar",
            )

        res = mocked.trigger(
            fn,
            inngest.Event(name="test"),
            client_mock,
            step_stubs={"a": "hi"},
        )
        assert res.status is mocked.Status.COMPLETED
        assert res.output == "hi"

    def test_parallel(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> tuple[str, ...]:
            return ctx.group.parallel(
                (
                    lambda: ctx.step.run("a", lambda: "a"),
                    lambda: ctx.step.run("b", lambda: "b"),
                )
            )

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.COMPLETED
        assert res.output == ("a", "b")

    def test_sleep(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> str:
            ctx.step.sleep("a", datetime.timedelta(seconds=1))
            return "hi"

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.COMPLETED
        assert res.output == "hi"

    def test_wait_for_event(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> typing.Mapping[str, inngest.JSON]:
            event = ctx.step.wait_for_event(
                "a",
                event="other-event",
                timeout=datetime.timedelta(seconds=1),
            )
            assert event is not None
            return event.data

        res = mocked.trigger(
            fn,
            inngest.Event(name="test"),
            client_mock,
            step_stubs={
                "a": inngest.Event(data={"foo": 1}, name="other-event")
            },
        )
        assert res.status is mocked.Status.COMPLETED
        assert res.output == {"foo": 1}

    def test_wait_for_event_timeout(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> None:
            event = ctx.step.wait_for_event(
                "a",
                event="other-event",
                timeout=datetime.timedelta(seconds=1),
            )
            assert event is None

        res = mocked.trigger(
            fn,
            inngest.Event(name="test"),
            client_mock,
            step_stubs={"a": mocked.Timeout},
        )
        assert res.status is mocked.Status.COMPLETED

    def test_wait_for_event_not_stubbed(self) -> None:
        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> None:
            ctx.step.wait_for_event(
                "a",
                event="other-event",
                timeout=datetime.timedelta(seconds=1),
            )

        with pytest.raises(UnstubbedStepError):
            mocked.trigger(fn, inngest.Event(name="test"), client_mock)

    def test_retry_step(self) -> None:
        counter = 0

        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> str:
            def a() -> str:
                nonlocal counter
                counter += 1
                if counter < 2:
                    raise Exception("oh no")
                return "hi"

            return ctx.step.run("a", a)

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.COMPLETED
        assert res.output == "hi"

    def test_fail_step(self) -> None:
        @client.create_function(
            fn_id="test",
            retries=0,
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> None:
            def a() -> None:
                raise Exception("oh no")

            ctx.step.run("a", a)

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.FAILED
        assert res.output is None
        assert isinstance(res.error, Exception)
        assert str(res.error) == "oh no"

    def test_retry_fn(self) -> None:
        counter = 0

        @client.create_function(
            fn_id="test",
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> str:
            nonlocal counter
            counter += 1
            if counter < 2:
                raise Exception("oh no")
            return "hi"

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.COMPLETED
        assert res.output == "hi"

    def test_fail_fn(self) -> None:
        @client.create_function(
            fn_id="test",
            retries=0,
            trigger=inngest.TriggerEvent(event="test"),
        )
        def fn(ctx: inngest.ContextSync) -> None:
            raise Exception("oh no")

        res = mocked.trigger(fn, inngest.Event(name="test"), client_mock)
        assert res.status is mocked.Status.FAILED
        assert res.output is None
        assert isinstance(res.error, Exception)
        assert str(res.error) == "oh no"



import datetime
import unittest.mock

import test_core

from .heartbeat_handler import _HeartbeatHandler
from .models import ConnectionState, _State
from .value_watcher import _ValueWatcher


class TestHeartbeatHandle(unittest.IsolatedAsyncioTestCase):
    async def test_send_error(self) -> None:
        # Count the number of times we send a heartbeat
        count_sends = 0

        # Create a mock WebSocket connection that always fails to send
        class _MockWS(unittest.mock.AsyncMock):
            async def send(self, *args: object, **kwargs: object) -> None:
                nonlocal count_sends
                count_sends += 1
                raise Exception("oh no")

        mock_ws = _MockWS()

        handler = _HeartbeatHandler(
            logger=unittest.mock.Mock(),
            state=_State(
                conn_id=None,
                conn_init=_ValueWatcher(None),
                conn_state=_ValueWatcher(ConnectionState.ACTIVE),
                exclude_gateways=[],
                extend_lease_interval=_ValueWatcher(None),
                fatal_error=_ValueWatcher(None),
                init_handshake_complete=_ValueWatcher(True),
                pending_request_count=_ValueWatcher(0),
                ws=_ValueWatcher(mock_ws),
            ),
        )
        handler.start()

        # Ensure that we send 2 heartbeats. This ensures that we continue
        # sending heartbeats even if one errors
        def assertion() -> None:
            assert count_sends == 2

        await test_core.wait_for(
            assertion, timeout=datetime.timedelta(seconds=10)
        )



import asyncio

from inngest._internal import types

from . import connect_pb2


class _BaseHandler:
    _closed_event: asyncio.Event | None = None

    @property
    def closed_event(self) -> asyncio.Event:
        if self._closed_event is None:
            self._closed_event = asyncio.Event()
        return self._closed_event

    def handle_msg(
        self,
        msg: connect_pb2.ConnectMessage,
        auth_data: connect_pb2.AuthData,
        connection_id: str,
    ) -> None:
        pass

    def start(self) -> types.MaybeError[None]:
        if self._closed_event is None:
            self._closed_event = asyncio.Event()
        return None

    def close(self) -> None:
        if self._closed_event is not None:
            self._closed_event.set()

    async def closed(self) -> None:
        if self._closed_event is not None:
            await self._closed_event.wait()



import asyncio
import typing
import urllib.parse

import httpx

from inngest._internal import comm_lib, net, server_lib, types

from . import connect_pb2, ws_utils
from .base_handler import _BaseHandler
from .buffer import _SizeConstrainedBuffer
from .models import _State
from .value_watcher import _ValueWatcher

# TODO: Make this configurable.
_default_max_buffer_size_bytes = 1024 * 1024 * 500  # 500MB


PendingRequest: typing.TypeAlias = tuple[
    connect_pb2.GatewayExecutorRequestData,
    asyncio.Task[None],
]


class _PendingRequestManager:
    def __init__(self, pending_request_count: _ValueWatcher[int]) -> None:
        self._pending_request_count = pending_request_count
        self._pending_requests: dict[str, PendingRequest] = {}

    def add(
        self,
        request_id: str,
        request_data: connect_pb2.GatewayExecutorRequestData,
        task: asyncio.Task[None],
    ) -> None:
        self._pending_requests[request_id] = (request_data, task)
        self._pending_request_count.value += 1

    def clear(self) -> None:
        self._pending_requests.clear()
        self._pending_request_count.value = 0

    def count(self) -> int:
        return len(self._pending_requests)

    def get(self, request_id: str) -> PendingRequest | None:
        return self._pending_requests.get(request_id, None)

    def get_all(
        self,
    ) -> list[PendingRequest]:
        return list(self._pending_requests.values())

    def pop(self, request_id: str) -> PendingRequest:
        req = self._pending_requests.pop(request_id)
        self._pending_request_count.value -= 1
        return req


class _ExecutionHandler(_BaseHandler):
    """
    Handles incoming execution requests from the Gateway.
    """

    _leaser_extender_task: asyncio.Task[None] | None = None
    _unacked_msg_flush_poller_task: asyncio.Task[None] | None = None

    def __init__(
        self,
        api_origin: str,
        comm_handlers: dict[str, comm_lib.CommHandler],
        http_client: net.ThreadAwareAsyncHTTPClient,
        http_client_sync: httpx.Client,
        logger: types.Logger,
        signing_key: str | None,
        signing_key_fallback: str | None,
        state: _State,
    ) -> None:
        self._api_origin = api_origin
        self._buffer = _SizeConstrainedBuffer(_default_max_buffer_size_bytes)
        self._comm_handlers = comm_handlers
        self._http_client = http_client
        self._http_client_sync = http_client_sync
        self._logger = logger
        self._signing_key = signing_key
        self._signing_key_fallback = signing_key_fallback
        self._state = state

        # Keep track of pending tasks to support graceful shutdown and lease
        # extensions.
        self._pending_requests = _PendingRequestManager(
            state.pending_request_count
        )

    def start(self) -> types.MaybeError[None]:
        err = super().start()
        if err is not None:
            return err

        if self._leaser_extender_task is None:
            self._leaser_extender_task = asyncio.create_task(
                self._leaser_extender()
            )

        if self._unacked_msg_flush_poller_task is None:
            self._unacked_msg_flush_poller_task = asyncio.create_task(
                self._unacked_msg_flush_poller()
            )

        return None

    def close(self) -> None:
        # Don't close until all pending requests end
        self._state.pending_request_count.on_value(0, super().close)

    async def closed(self) -> None:
        await super().closed()
        await asyncio.gather(*[t for _, t in self._pending_requests.get_all()])
        self._pending_requests.clear()

        if self._leaser_extender_task is not None:
            # Force lease extension to stop
            self._leaser_extender_task.cancel()
            try:
                await self._leaser_extender_task
            except asyncio.CancelledError:
                # This is expected since the task is likely calling
                # `asyncio.sleep` after cancellation.
                pass

        if self._unacked_msg_flush_poller_task is not None:
            # Force unacked message flush to stop
            self._unacked_msg_flush_poller_task.cancel()
            try:
                await self._unacked_msg_flush_poller_task
            except asyncio.CancelledError:
                # This is expected since the task is likely calling
                # `asyncio.sleep` after cancellation.
                pass

    def handle_msg(
        self,
        msg: connect_pb2.ConnectMessage,
        auth_data: connect_pb2.AuthData,
        connection_id: str,
    ) -> None:
        if msg.kind == connect_pb2.GatewayMessageType.GATEWAY_EXECUTOR_REQUEST:
            self._handle_executor_request(msg)
        elif (
            msg.kind
            == connect_pb2.GatewayMessageType.WORKER_REQUEST_EXTEND_LEASE_ACK
        ):
            self._handle_lease_extend_ack(msg)
        elif msg.kind == connect_pb2.GatewayMessageType.WORKER_REPLY_ACK:
            self._handle_worker_reply_ack(msg)

    def _handle_executor_request(
        self,
        msg: connect_pb2.ConnectMessage,
    ) -> None:
        self._logger.debug("Received executor request")

        req_data = connect_pb2.GatewayExecutorRequestData()
        req_data.ParseFromString(msg.payload)

        comm_handler = self._comm_handlers.get(req_data.app_name)
        if comm_handler is None:
            self._logger.error(
                "Unknown app",
                extra={"app_id": req_data.app_name},
            )
            return

        async def execute() -> None:
            ws = await self._state.ws.wait_for_not_none()
            err = await ws_utils.safe_send(
                self._logger,
                self._state,
                ws,
                connect_pb2.ConnectMessage(
                    kind=connect_pb2.GatewayMessageType.WORKER_REQUEST_ACK,
                    payload=connect_pb2.WorkerRequestAckData(
                        account_id=req_data.account_id,
                        app_id=req_data.app_id,
                        env_id=req_data.env_id,
                        function_slug=req_data.function_slug,
                        request_id=req_data.request_id,
                        step_id=req_data.step_id,
                        system_trace_ctx=req_data.system_trace_ctx,
                        user_trace_ctx=req_data.user_trace_ctx,
                    ).SerializeToString(),
                ).SerializeToString(),
            )
            if err is None:
                comm_res = await comm_handler.post(
                    comm_lib.CommRequest(
                        body=req_data.request_payload,
                        headers={},
                        is_connect=True,
                        public_path=None,
                        query_params={
                            server_lib.QueryParamKey.FUNCTION_ID.value: req_data.function_slug,
                        },
                        raw_request=req_data,
                        request_url="",
                        serve_origin=None,
                        serve_path=None,
                    )
                )
            else:
                self._logger.error(
                    "Execution failed", extra={"error": str(err)}
                )
                comm_res = comm_lib.CommResponse.from_error(self._logger, err)

            body = comm_res.body_bytes()
            if isinstance(body, Exception):
                raise body

            status: connect_pb2.SDKResponseStatus
            if comm_res.status_code == 200:
                # Function-level success.
                status = connect_pb2.SDKResponseStatus.DONE
            elif comm_res.status_code == 206:
                # Step-level success.
                status = connect_pb2.SDKResponseStatus.NOT_COMPLETED
            elif comm_res.status_code == 500:
                # Error.
                status = connect_pb2.SDKResponseStatus.ERROR
            else:
                # Unreachable.
                status = connect_pb2.SDKResponseStatus.ERROR
                self._logger.error(
                    "Unexpected status code",
                    extra={"status_code": comm_res.status_code},
                )

            reply_payload = connect_pb2.SDKResponse(
                account_id=req_data.account_id,
                app_id=req_data.app_id,
                body=body,
                env_id=req_data.env_id,
                no_retry=comm_res.no_retry,
                request_id=req_data.request_id,
                request_version=comm_res.request_version,
                retry_after=comm_res.retry_after,
                sdk_version=comm_res.sdk_version,
                status=status,
                system_trace_ctx=req_data.system_trace_ctx,
                user_trace_ctx=req_data.user_trace_ctx,
            ).SerializeToString()

            # Add the reply to the buffer in case we need to flush it later.
            self._buffer.add(req_data.request_id, reply_payload)

            self._logger.debug("Sending execution reply")
            err = await ws_utils.safe_send(
                self._logger,
                self._state,
                ws,
                connect_pb2.ConnectMessage(
                    kind=connect_pb2.GatewayMessageType.WORKER_REPLY,
                    payload=reply_payload,
                ).SerializeToString(),
            )
            if err is not None:
                self._logger.error(
                    "Failed to send execution reply", extra={"error": str(err)}
                )

        # Store the task.
        task = asyncio.create_task(execute())
        self._pending_requests.add(req_data.request_id, req_data, task)

        # Remove the task when it completes.
        task.add_done_callback(
            lambda _: self._pending_requests.pop(req_data.request_id)
        )

    def _handle_lease_extend_ack(
        self,
        msg: connect_pb2.ConnectMessage,
    ) -> None:
        req_data = connect_pb2.WorkerRequestExtendLeaseAckData()
        req_data.ParseFromString(msg.payload)

        pending_req = self._pending_requests.get(req_data.request_id)
        if pending_req is None:
            self._logger.error(
                "Received lease extend ack for unknown request",
                extra={"request_id": req_data.request_id},
            )
            return

        if req_data.new_lease_id:
            # Each lease extension ack includes a new lease ID. If we don't use the
            # new lease ID the next time we extend, we'll have a bad time.
            pending_req[0].lease_id = req_data.new_lease_id
        else:
            # A null new_lease_id indicates that the lease extension failed. This can happen
            # if the lease was expired, deleted, or taken over by another worker, so we should
            # stop trying to extend it.
            self._logger.debug(
                "Unable to extend lease",
                extra={"request_id": req_data.request_id},
            )
            # Cancelling the task will trigger the done callback to remove it from
            # the pending requests.
            pending_req[1].cancel()

    def _handle_worker_reply_ack(
        self,
        msg: connect_pb2.ConnectMessage,
    ) -> None:
        """
        Handles a worker reply ack, which indicates that the Inngest Server
        received the execution reply.
        """

        req_data = connect_pb2.WorkerReplyAckData()
        req_data.ParseFromString(msg.payload)

        self._logger.debug(
            "Received worker reply ack",
            extra={"request_id": req_data.request_id},
        )
        self._buffer.delete(req_data.request_id)

    async def _leaser_extender(self) -> None:
        while self.closed_event.is_set() is False:
            ws = await self._state.ws.wait_for_not_none()
            extend_lease_interval = (
                await self._state.extend_lease_interval.wait_for_not_none()
            )

            await asyncio.sleep(extend_lease_interval)

            if self._pending_requests.count() == 0:
                continue

            self._logger.debug(
                "Extending leases",
                extra={"count": self._pending_requests.count()},
            )

            for req_data, _ in self._pending_requests.get_all():
                err = await ws_utils.safe_send(
                    self._logger,
                    self._state,
                    ws,
                    connect_pb2.ConnectMessage(
                        kind=connect_pb2.GatewayMessageType.WORKER_REQUEST_EXTEND_LEASE,
                        payload=connect_pb2.WorkerRequestExtendLeaseData(
                            account_id=req_data.account_id,
                            env_id=req_data.env_id,
                            function_slug=req_data.function_slug,
                            lease_id=req_data.lease_id,
                            request_id=req_data.request_id,
                            run_id=req_data.run_id,
                            step_id=req_data.step_id,
                            system_trace_ctx=req_data.system_trace_ctx,
                            user_trace_ctx=req_data.user_trace_ctx,
                        ).SerializeToString(),
                    ).SerializeToString(),
                )
                if err is not None:
                    self._logger.error(
                        "Failed to extend lease", extra={"error": str(err)}
                    )

    async def _unacked_msg_flush_poller(self) -> None:
        """
        Responsible for flushing unacked messages via HTTP. Checks for flushable
        messages on an interval.
        """

        # How long a message should exist before being flushed. We picked the
        # lease interval, but there may be a better value.
        flush_ttl = await self._state.extend_lease_interval.wait_for_not_none()

        while self.closed_event.is_set() is False:
            for request_id, reply_msg in self._buffer.get_older_than(flush_ttl):
                try:
                    err = await self._flush_message(reply_msg)
                    if err is not None:
                        self._logger.error(
                            "Failed to flush message", extra={"error": str(err)}
                        )
                finally:
                    # We only attempt to flush once, so we can delete the
                    # message.
                    self._buffer.delete(request_id)

            await asyncio.sleep(1)

    async def _flush_message(self, msg: bytes) -> types.MaybeError[None]:
        """
        Flush a single message via HTTP.
        """

        url = urllib.parse.urljoin(self._api_origin, "/v0/connect/flush")

        res = await net.fetch_with_auth_fallback(
            self._http_client,
            self._http_client_sync,
            httpx.Request(
                content=msg,
                extensions={
                    "timeout": httpx.Timeout(5).as_dict(),
                },
                headers={
                    "content-type": "application/protobuf",
                },
                method="POST",
                url=url,
            ),
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if isinstance(res, Exception):
            return res
        if res.status_code == 401 or res.status_code == 403:
            return Exception("unauthorized")
        if res.status_code < 200 or res.status_code >= 300:
            return Exception(f"unexpected status code: {res.status_code}")

        return None



from typing_extensions import assert_type

from . import types


class Test_is_dict:
    def test_dict(self) -> None:
        v: dict[object, object] = {}
        assert types.is_dict(v)
        assert_type(v, dict[object, object])

    def test_non_dicts(self) -> None:
        assert types.is_dict([]) is False
        assert types.is_dict("string") is False
        assert types.is_dict(123) is False
        assert types.is_dict(None) is False
        assert types.is_dict(set[str]()) is False
        assert types.is_dict(tuple[str]()) is False


class Test_is_dict_keys:
    def test_key_type(self) -> None:
        v: dict[object, object] = {"a": 1}
        assert types.is_dict_keys(v, str)
        assert_type(v, dict[str, object])

    def test_key_type_invalid(self) -> None:
        v: dict[object, object] = {1: "a"}
        assert types.is_dict_keys(v, str) is False


class Test_is_list:
    def test_list(self) -> None:
        v: list[object] = []
        assert types.is_list(v)
        assert_type(v, list[object])

    def test_non_lists(self) -> None:
        assert types.is_list({}) is False
        assert types.is_list("string") is False
        assert types.is_list(123) is False
        assert types.is_list(None) is False
        assert types.is_list(set[str]()) is False
        assert types.is_list(tuple[str]()) is False


class Test_is_list_items:
    def test_item_type_primitive(self) -> None:
        v: list[object] = ["a"]
        assert types.is_list_items(v, str)
        assert_type(v, list[str])

    def test_item_type_object(self) -> None:
        class Foo:
            pass

        v: list[Foo] = []
        assert types.is_list_items(v, Foo)
        assert_type(v, list[Foo])

    def test_item_type_invalid(self) -> None:
        v: list[object] = [1]
        assert types.is_list_items(v, str) is False



from __future__ import annotations

import datetime
import logging
import unittest

import inngest
from inngest._internal import errors

from .handler import get_function_configs

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
client = inngest.Inngest(
    api_base_url="http://foo.bar",
    app_id="test",
    is_production=False,
    logger=logger,
)


class Test_get_function_configs(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)

    def test_full_config(self) -> None:
        """
        Ensure that there isn't a validation error when creating a
        fully-specified config.
        """

        @client.create_function(
            batch_events=inngest.Batch(
                max_size=2, timeout=datetime.timedelta(minutes=1)
            ),
            cancel=[
                inngest.Cancel(
                    event="app/cancel",
                    if_exp="true",
                    timeout=datetime.timedelta(minutes=1),
                )
            ],
            fn_id="fn",
            name="Function",
            retries=1,
            throttle=inngest.Throttle(
                limit=2, period=datetime.timedelta(minutes=1)
            ),
            singleton=inngest.Singleton(key="false", mode="skip"),
            trigger=inngest.TriggerEvent(event="app/fn"),
        )
        def fn(ctx: inngest.ContextSync) -> int:
            return 1

        configs = get_function_configs("http://foo.bar", {fn.id: fn})
        assert not isinstance(configs, Exception), (
            f"Unexpected error: {configs}"
        )

    def test_no_functions(self) -> None:
        configs = get_function_configs("http://foo.bar", {})
        assert isinstance(configs, errors.FunctionConfigInvalidError)
        assert str(configs) == "no functions found"



import signal

import websockets

from inngest._internal import server_lib

_default_shutdown_signals = [signal.SIGINT, signal.SIGTERM]
_framework = server_lib.Framework.CONNECT
_heartbeat_interval_sec = 10
_protocol = websockets.Subprotocol("v0.connect.inngest.com")



from __future__ import annotations

import datetime
import typing

import pydantic

from inngest._internal import server_lib, transforms


class UnreachableError(Exception):
    """
    We should never raise this error. Its only purpose is to explicitly indicate
    theoretically unreachable code.
    """

    pass


class Error(Exception):
    """
    Base error for all our custom errors
    """

    code: server_lib.ErrorCode = server_lib.ErrorCode.UNKNOWN
    include_stack: bool = True
    is_retriable: bool = True

    @property
    def message(self) -> str:
        return str(self)

    @property
    def name(self) -> str:
        return type(self).__name__

    @property
    def stack(self) -> str | None:
        if self.include_stack is False:
            return None

        return transforms.get_traceback(self)


class CodedError(Error):
    def __init__(
        self,
        code: server_lib.ErrorCode,
        message: str,
        *,
        is_retriable: bool = True,
    ) -> None:
        super().__init__(message)
        self.code = code
        self.is_retriable = is_retriable


class URLInvalidError(CodedError):
    code = server_lib.ErrorCode.URL_INVALID


class FunctionConfigInvalidError(Error):
    code = server_lib.ErrorCode.FUNCTION_CONFIG_INVALID

    @classmethod
    def from_validation_error(
        cls,
        err: pydantic.ValidationError,
    ) -> FunctionConfigInvalidError:
        """
        Extract info from Pydantic's ValidationError and return our internal
        InvalidFunctionConfig error.
        """
        default = cls(str(err))

        errors = err.errors()
        if len(errors) == 0:
            return default
        loc = errors[0].get("loc")
        if loc is None or len(loc) == 0:
            return default

        field = ""
        for part in loc:
            if isinstance(part, int):
                return default
            if len(field) > 0:
                field += "."
            field += part

        msg = errors[0].get("msg")
        if msg is None:
            return default

        return cls(f"{loc[0]}: {msg}")


class AsyncUnsupportedError(Error):
    code = server_lib.ErrorCode.ASYNC_UNSUPPORTED


class SigVerificationFailedError(Error):
    code = server_lib.ErrorCode.SIG_VERIFICATION_FAILED


class BodyInvalidError(Error):
    code = server_lib.ErrorCode.BODY_INVALID


class EventKeyUnspecifiedError(Error):
    code = server_lib.ErrorCode.EVENT_KEY_UNSPECIFIED


class FunctionNotFoundError(Error):
    code = server_lib.ErrorCode.FUNCTION_NOT_FOUND


class HeaderMissingError(Error):
    code = server_lib.ErrorCode.HEADER_MISSING


class QueryParamMissingError(Error):
    code = server_lib.ErrorCode.QUERY_PARAM_MISSING


class SigningKeyMissingError(Error):
    code = server_lib.ErrorCode.SIGNING_KEY_UNSPECIFIED


class RegistrationFailedError(Error):
    code = server_lib.ErrorCode.REGISTRATION_FAILED


class UnknownError(Error):
    code = server_lib.ErrorCode.UNKNOWN


class StepUnexpectedError(Error):
    code = server_lib.ErrorCode.STEP_UNEXPECTED
    include_stack = False


class OutputUnserializableError(Error):
    code = server_lib.ErrorCode.OUTPUT_UNSERIALIZABLE
    is_retriable = False


class NonRetriableError(Error):
    """End users can raise this error to prevent retries."""

    code = server_lib.ErrorCode.NON_RETRIABLE_ERROR
    is_retriable = False

    def __init__(
        self,
        message: str | None = None,
        quiet: bool = False,
    ) -> None:
        super().__init__(message)
        self.quiet = quiet


class RetryAfterError(Error):
    code = server_lib.ErrorCode.RETRY_AFTER_ERROR

    def __init__(
        self,
        message: str | None,
        retry_after: int | datetime.timedelta | datetime.datetime,
        quiet: bool = False,
    ) -> None:
        """
        Raise this error to retry at a specific time.

        Args:
        ----
            message: Error message
            retry_after: Time to retry after in milliseconds, timedelta, or datetime
            quiet: Whether to supress logging
        """

        super().__init__(message)

        if isinstance(retry_after, int):
            retry_after = datetime.datetime.now() + datetime.timedelta(
                milliseconds=retry_after
            )
        elif isinstance(retry_after, datetime.timedelta):
            retry_after = datetime.datetime.now() + retry_after

        self.retry_after: datetime.datetime = retry_after
        self.quiet: bool = quiet


class SendEventsError(Error):
    code = server_lib.ErrorCode.SEND_EVENT_FAILED

    def __init__(self, message: str, ids: list[str]) -> None:
        """
        Args:
        ----
            message: Error message
            ids: List of event IDs that successfully sent
        """

        super().__init__(message)
        self.ids = ids


class StepError(Error):
    """
    Wraps a userland error. This is necessary because the Executor sends
    memoized error data which can't be deserialized into the original error
    class.
    """

    code = server_lib.ErrorCode.STEP_ERRORED

    # Not retriable since this error is thrown after exhausting retries
    is_retriable = False

    @property
    def message(self) -> str:
        """
        Returns the userland error message
        """

        return self._message

    @property
    def name(self) -> str:
        """
        Returns the userland error name
        """

        return self._name

    @property
    def stack(self) -> str | None:
        """
        Returns the userland error stack trace
        """

        return self._stack

    def __init__(
        self,
        message: str,
        name: str,
        stack: str | None,
    ) -> None:
        """
        Args:
        ----
            message: Userland error's message
            name: Userland error's name
            stack: Userland error's stack trace
        """

        super().__init__(message)
        self._message = message
        self._name = name
        self._stack = stack


def is_retriable(err: Exception) -> bool:
    if isinstance(err, Error):
        return err.is_retriable
    return True


def is_quiet(err: Exception) -> bool:
    if isinstance(err, _Quietable):
        return err.quiet
    return False


@typing.runtime_checkable
class _Quietable(typing.Protocol):
    quiet: bool



from __future__ import annotations

import dataclasses
import threading
import typing

import pydantic

from inngest._internal import (
    client_lib,
    middleware_lib,
    server_lib,
    transforms,
    types,
)


class MemoizedError(types.BaseModel):
    message: str
    name: str
    stack: str | None = None

    @classmethod
    def from_error(cls, err: Exception) -> MemoizedError:
        return cls(
            message=str(err),
            name=type(err).__name__,
            stack=transforms.get_traceback(err),
        )


class Output(types.BaseModel):
    # Fail validation if any extra fields exist, because this will prevent
    # accidentally assuming user data is nested data
    model_config = pydantic.ConfigDict(extra="forbid")

    data: object = None
    error: MemoizedError | None = None


class StepMemos:
    """Holds memoized step output."""

    @property
    def size(self) -> int:
        return len(self._memos)

    def __init__(self, memos: dict[str, Output]) -> None:
        self._memos = memos

    def values(self) -> typing.Iterator[Output]:
        return iter(self._memos.values())

    def pop(self, hashed_id: str) -> Output | types.EmptySentinel:
        if hashed_id in self._memos:
            memo = self._memos[hashed_id]

            # Remove memo
            self._memos = {
                k: v for k, v in self._memos.items() if k != hashed_id
            }

            return memo

        return types.empty_sentinel

    @classmethod
    def from_raw(cls, raw: dict[str, object]) -> StepMemos:
        memos = {}
        for k, v in raw.items():
            output = Output.from_raw(v)
            if isinstance(output, Exception):
                # Not all steps nest their output in an Output-compatible object
                # (i.e. they don't nest output in a data field). For example,
                # `step.run` nests its output in a data field but
                # `step.waitForEvent` will not nest its fulfilling event.
                output = Output(data=v)

            memos[k] = output
        return cls(memos)


class StepBase:
    def __init__(
        self,
        client: client_lib.Inngest,
        middleware: middleware_lib.MiddlewareManager,
        step_id_counter: StepIDCounter,
        target_hashed_id: str | None,
    ) -> None:
        self._client = client
        self._middleware = middleware
        self._step_id_counter = step_id_counter
        self._target_hashed_id = target_hashed_id

    def _handle_skip(
        self,
        parsed_step_id: ParsedStepID,
    ) -> None:
        """
        Handle a skip interrupt. Step targeting is enabled and this step is not
        the target then skip the step.
        """

        is_targeting_enabled = self._target_hashed_id is not None
        is_targeted = self._target_hashed_id == parsed_step_id.hashed
        if is_targeting_enabled and not is_targeted:
            # Skip this step because a different step is targeted.
            raise SkipInterrupt(parsed_step_id.user_facing)

    def _parse_step_id(self, step_id: str) -> ParsedStepID:
        """
        Parse a user-specified step ID into a hashed ID and a deduped
        user-facing step ID.
        """

        pre_hashed = step_id
        id_count = self._step_id_counter.increment(step_id)
        if id_count > 1:
            pre_hashed = f"{step_id}:{id_count - 1}"

        return ParsedStepID(
            hashed=transforms.hash_step_id(pre_hashed),
            user_facing=step_id,
            index=(id_count - 1) if id_count > 1 else None,
        )


@dataclasses.dataclass
class ParsedStepID:
    hashed: str
    user_facing: str
    index: int | None

    def userland_info(self) -> StepUserlandInfo:
        return StepUserlandInfo(
            id=self.user_facing,
            index=self.index,
        )


class StepIDCounter:
    """
    Counts the number of times a step ID has been used. We support reused step
    IDs so we need a way to keep track of their counts.
    """

    def __init__(self) -> None:
        self._counts: dict[str, int] = {}
        self._mutex = threading.Lock()

    def increment(self, hashed_id: str) -> int:
        with self._mutex:
            if hashed_id not in self._counts:
                self._counts[hashed_id] = 0

            self._counts[hashed_id] += 1
            return self._counts[hashed_id]


class ResponseInterrupt(BaseException):
    """
    Extend BaseException to avoid being caught by the user's code. Users can
    still catch it if they do a "bare except", but that's a known antipattern in
    the Python world.
    """

    def __init__(
        self,
        responses: StepResponse | list[StepResponse],
    ) -> None:
        if not isinstance(responses, list):
            responses = [responses]
        self.responses = responses

    def set_step_info(self, step_info: StepInfo) -> None:
        """
        Set the step info on all responses. This is useful for catching nested
        steps, since we want to report the "parent" step info and not the
        "child" step info.
        """

        for response in self.responses:
            response.step = step_info


class SkipInterrupt(BaseException):
    def __init__(self, step_id: str) -> None:
        self.step_id = step_id


class NestedStepInterrupt(BaseException):
    pass


class InvokeOpts(types.BaseModel):
    function_id: str
    payload: InvokeOptsPayload
    timeout: str | None


class InvokeOptsPayload(types.BaseModel):
    data: object
    v: str | None


class WaitForEventOpts(types.BaseModel):
    if_exp: str | None = pydantic.Field(..., serialization_alias="if")
    timeout: str


class AIInferOpts(types.BaseModel):
    auth_key: str
    body: dict[str, object]
    format: str
    headers: dict[str, str]
    type: str = "step.ai.infer"
    url: str | None = None


class StepInfo(types.BaseModel):
    display_name: str = pydantic.Field(..., serialization_alias="displayName")
    id: str

    # Deprecated
    name: str | None = None

    op: server_lib.Opcode
    opts: dict[str, object] | None = None
    userland: StepUserlandInfo | None = None

    def set_parallel_mode(self, parallel_mode: server_lib.ParallelMode) -> None:
        if parallel_mode != server_lib.ParallelMode.RACE:
            # Nothing to do because race mode is opt-in
            return

        if self.opts is None:
            self.opts = {}
        self.opts[server_lib.OptKey.PARALLEL_MODE.value] = parallel_mode.value


@dataclasses.dataclass
class StepUserlandInfo:
    id: str
    index: int | None


class StepResponse(types.BaseModel):
    output: object = None
    original_error: object = pydantic.Field(default=None, exclude=True)
    step: StepInfo



from __future__ import annotations

from . import types


class LoggerProxy:
    """
    Wraps a logger, allowing us to disable logging when we want to. This is
    important because we may call a function multiple times and we don't want
    duplicate logs.
    """

    _proxied_methods = (
        "critical",
        "debug",
        "error",
        "exception",
        "fatal",
        "info",
        "log",
        "warn",
        "warning",
    )

    def __init__(self, logger: types.Logger) -> None:
        self._is_enabled = False
        self.logger = logger

    def __getattr__(self, name: str) -> object:
        if name in self._proxied_methods and not self._is_enabled:
            # Return noop
            return lambda *args, **kwargs: None

        return getattr(self.logger, name)

    def enable(self) -> None:
        self._is_enabled = True



"""
Simulate Inngest function execution without an Inngest server.

NOT STABLE! This is an experimental feature and may change in the future. If
you'd like to depend on it, we recommend copying this directory into your source
code.
"""

from .client import Inngest
from .consts import Status, Timeout
from .trigger import trigger

__all__ = [
    "Inngest",
    "Status",
    "Timeout",
    "trigger",
]



import asyncio

from inngest._internal import types

from . import connect_pb2, ws_utils
from .base_handler import _BaseHandler
from .consts import _heartbeat_interval_sec
from .models import _State


class _HeartbeatHandler(_BaseHandler):
    """
    Sends outgoing heartbeats to the Gateway and handles incoming heartbeat
    messages.
    """

    _heartbeat_sender_task: asyncio.Task[None] | None = None

    def __init__(
        self,
        logger: types.Logger,
        state: _State,
    ) -> None:
        self._logger = logger
        self._state = state

    def start(self) -> types.MaybeError[None]:
        err = super().start()
        if err is not None:
            return err

        if self._heartbeat_sender_task is None:
            self._heartbeat_sender_task = asyncio.create_task(
                self._heartbeat_sender()
            )

        return None

    def close(self) -> None:
        self._state.pending_request_count.on_value(0, super().close)

    async def closed(self) -> None:
        await super().closed()
        await self._state.pending_request_count.wait_for(0)

        if self._heartbeat_sender_task is not None:
            self._heartbeat_sender_task.cancel()
            try:
                await self._heartbeat_sender_task
            except asyncio.CancelledError:
                # This is expected since the task is likely calling
                # `asyncio.sleep` after cancellation.
                pass

    async def _heartbeat_sender(
        self,
    ) -> None:
        while self.closed_event.is_set() is False:
            # We can't send heartbeats until the handshake is complete. Doing so
            # isn't fatal, but will result in "connect_worker_hello_invalid_msg"
            # logs
            await self._state.init_handshake_complete.wait_for(True)

            # IMPORTANT: We need to get the WS conn each loop iteration because
            # it may have changed (e.g. due to a reconnect)
            ws = await self._state.ws.wait_for_not_none()

            self._logger.debug("Sending heartbeat")
            err = await ws_utils.safe_send(
                self._logger,
                self._state,
                ws,
                connect_pb2.ConnectMessage(
                    kind=connect_pb2.GatewayMessageType.WORKER_HEARTBEAT,
                ).SerializeToString(),
            )
            if err is not None:
                # Only log the error because we want to continue heartbeating
                self._logger.error(
                    "Error sending heartbeat", extra={"error": str(err)}
                )

            await asyncio.sleep(_heartbeat_interval_sec)

        self._logger.debug("Heartbeater task stopped")

    def handle_msg(
        self,
        msg: connect_pb2.ConnectMessage,
        auth_data: connect_pb2.AuthData,
        connection_id: str,
    ) -> None:
        if msg.kind != connect_pb2.GatewayMessageType.GATEWAY_HEARTBEAT:
            return

        self._logger.debug("Received heartbeat")
